'''
Module providing subsidary functions for tensor mapping in 3D
'''

from numpy import \
    array, zeros, ix_, hstack, vstack

import numpy as np


#-------------------------------------------------------------------------
# Switch from engineering to tensorial notation
#-------------------------------------------------------------------------
def map3d_eps_eng_to_mtx(eps_eng):
    '''
    Switch from engineering notation to tensor notation for strains in 3D
    '''
    eps_mtx = array([[eps_eng[0], eps_eng[5] / 2., eps_eng[4] / 2.],
                     [eps_eng[5] / 2., eps_eng[1], eps_eng[3] / 2.],
                     [eps_eng[4] / 2., eps_eng[3] / 2, eps_eng[2]]])
    return eps_mtx


def map3d_sig_eng_to_mtx(sig_eng):
    '''
    Switch from engineering notation to tensor notation for stresses in 3D
    '''
    sig_mtx = array([[sig_eng[0], sig_eng[5], sig_eng[4]],
                     [sig_eng[5], sig_eng[1], sig_eng[3]],
                     [sig_eng[4], sig_eng[3], sig_eng[2]]])
    return sig_mtx


#-------------------------------------------------------------------------
# Switch from tensorial to engineering notation
#-------------------------------------------------------------------------

def map3d_eps_mtx_to_eng(eps_mtx):
    '''
    Switch from tensor notation to engineering notation for strains in 3D
    '''
    I = np.array([[0, 0], [1, 1], [2, 2]])
    eps_eng = array([eps_mtx[0, 0],
                     eps_mtx[1, 1],
                     eps_mtx[2, 2],
                     2.0 * eps_mtx[1, 2],
                     2.0 * eps_mtx[0, 2],
                     2.0 * eps_mtx[0, 1]])
    return eps_eng


def map3d_sig_mtx_to_eng(sig_mtx):
    '''
    Switch from tensor notation to engineering notation for stresses in 3D
    '''
    sig_eng = array([sig_mtx[0, 0],
                     sig_mtx[1, 1],
                     sig_mtx[2, 2],
                     sig_mtx[1, 2],
                     sig_mtx[0, 2],
                     sig_mtx[0, 1]])
    return sig_eng

#-------------------------------------------------------------------------
# Subsidiary index mapping functions for rank-four to rank-two tensors
#-------------------------------------------------------------------------


def map3d_ijkl2mn(i, j, k, l):
    '''
    Map the four-rank indexes to the two-rank matrix using the major
    and minor symmetry.
    '''
    # 3D-case:
    # first two indices (ij)
    if i == 0 and j == 0:
        m = 0
    elif i == 1 and j == 1:
        m = 1
    elif i == 2 and j == 2:
        m = 2
    elif (i == 1 and j == 2) or (i == 2 and j == 1):
        m = 3
    elif (i == 0 and j == 2) or (i == 2 and j == 0):
        m = 4
    elif (i == 0 and j == 1) or (i == 1 and j == 0):
        m = 5
    else:
        raise IndexError('error in the tensor index mapping')

    # second two indices (kl)
    if k == 0 and l == 0:
        n = 0
    elif k == 1 and l == 1:
        n = 1
    elif k == 2 and l == 2:
        n = 2
    elif (k == 1 and l == 2) or (k == 2 and l == 1):
        n = 3
    elif (k == 0 and l == 2) or (k == 2 and l == 0):
        n = 4
    elif (k == 0 and l == 1) or (k == 1 and l == 0):
        n = 5
    else:
        raise IndexError('error in the tensor index mapping')

    return m, n

#-------------------------------------------------------------------------
# Subsidiary mapping functions for rank-two to rank-four tensor
#-------------------------------------------------------------------------


def map3d_tns2_to_tns4(tns2):
    '''
    Map a matrix to a fourth order tensor assuming minor and major symmetry,
    e.g. D_mtx (6x6) in engineering notation to D_tns(3,3,3,3)).
    '''
    n_dim = 3
    tns4 = zeros([n_dim, n_dim, n_dim, n_dim])
    for i in range(0, n_dim):
        for j in range(0, n_dim):
            for k in range(0, n_dim):
                for l in range(0, n_dim):
                    tns4[i, j, k, l] = tns2[map3d_ijkl2mn(i, j, k, l)]
    return tns4


#-------------------------------------------------------------------------
# Subsidiary mapping functions for rank-four to rank-two tensor
#-------------------------------------------------------------------------

def map3d_tns4_to_tns2(tns4):
    '''
    Map a fourth order tensor to a matrix assuming minor and major symmetry,
    e.g. D_tns(3,3,3,3) to D_mtx (6x6) in engineering notation.
    (Note: Explicit assignment of components used for speed-up.)
    '''
    print('XXXXXXXXXXXXXXXXXXXXXXXXX')
    n_eng = 6
    tns2 = zeros([n_eng, n_eng])

    tns2[0, 0] = tns4[0, 0, 0, 0]
    tns2[0, 1] = tns2[1, 0] = tns4[0, 0, 1, 1]
    tns2[0, 2] = tns2[2, 0] = tns4[0, 0, 2, 2]
    tns2[0, 3] = tns2[3, 0] = tns4[0, 0, 1, 2]
    tns2[0, 4] = tns2[4, 0] = tns4[0, 0, 0, 2]
    tns2[0, 5] = tns2[5, 0] = tns4[0, 0, 0, 1]

    tns2[1, 1] = tns4[1, 1, 1, 1]
    tns2[1, 2] = tns2[2, 1] = tns4[1, 1, 2, 2]
    tns2[1, 3] = tns2[3, 1] = tns4[1, 1, 1, 2]
    tns2[1, 4] = tns2[4, 1] = tns4[1, 1, 0, 2]
    tns2[1, 5] = tns2[5, 1] = tns4[1, 1, 0, 1]

    tns2[2, 2] = tns4[2, 2, 2, 2]
    tns2[2, 3] = tns2[3, 2] = tns4[2, 2, 1, 2]
    tns2[2, 4] = tns2[4, 2] = tns4[2, 2, 0, 2]
    tns2[2, 5] = tns2[5, 2] = tns4[2, 2, 0, 1]

    tns2[3, 3] = tns4[1, 2, 1, 2]
    tns2[3, 4] = tns2[4, 3] = tns4[1, 2, 0, 2]
    tns2[3, 5] = tns2[5, 3] = tns4[1, 2, 0, 1]

    tns2[4, 4] = tns4[0, 2, 0, 2]
    tns2[4, 5] = tns2[5, 4] = tns4[0, 2, 0, 1]

    tns2[5, 5] = tns4[0, 1, 0, 1]

    return tns2


#-------------------------------------------------------------------------
# Compliance mapping 3D (used for inversion of the damage effect tensor in engineering notation
#-------------------------------------------------------------------------

def compliance_mapping3d(C_mtx_3d):
    '''
    The components of the compliance matrix are multiplied with factor 1,2 or 4 depending on their 
    position in the matrix (due to symmetry and switching of tensorial to engineering notation
    (Note: gamma_xy = 2*epsilon_xy, etc.). Necessary for evaluation of D=inv(C). 
    '''
    idx1 = [0, 1, 2]
    idx2 = [3, 4, 5]
    C11 = C_mtx_3d[ix_(idx1, idx1)]
    C12 = C_mtx_3d[ix_(idx1, idx2)]
    C21 = C_mtx_3d[ix_(idx2, idx1)]
    C22 = C_mtx_3d[ix_(idx2, idx2)]
    return vstack([hstack([C11,   2 * C12]),
                   hstack([2 * C21, 4 * C22])])



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 3, 2009 by jakub

from traits.api import Callable, Constant
from ibvpy.tmodel.mats_eval import MATSEval

def identity_mapping( var ):
    return var

def reshape( var ):
    return var.reshape( ( 1, 1 ) )

def flatten( var ):
    return var.flatten()

class MATS1DEval( MATSEval ):

    n_dims = Constant( 1 )

    # dimension-dependent mappings
    #
    map_tns4_to_tns2 = Callable( identity_mapping, transient = True )
    map_eps_eng_to_mtx = Callable( reshape, transient = True )
    map_sig_eng_to_mtx = Callable( reshape, transient = True )
    compliance_mapping = Callable( identity_mapping, transient = True )
    map_sig_mtx_to_eng = Callable( flatten, transient = True )
    map_eps_mtx_to_eng = Callable( flatten, transient = True )




from traits.api import \
    Property, cached_property, \
    provides, Array

from ibvpy.mathkit import EPS
import numpy as np
from ibvpy.sim.i_xmodel import IXModel

from .xdomain_fe_grid import XDomainFEGrid


@provides(IXModel)
class XDomainFEGridTransform(XDomainFEGrid):

    vtk_expand_operator = Array(np.float64)

    def _vtk_expand_operator_default(self):
        return np.identity(3)

    T_Emra = Property(depends_on='MESH,GEO,CS,FE')

    @cached_property
    def _get_T_Emra(self):
        To3D = self.fets.vtk_expand_operator
        J_Emar = np.einsum(
            'ik,...ij,jl->...kl',
            To3D, self.J_Emar, To3D
        )
        m_0_Ema = J_Emar[..., 0]

        m_2_Ema = np.einsum(
            '...i,...j,ijk->...k',
            m_0_Ema, J_Emar[..., 1], EPS
        )
        m_1_Ema = np.einsum(
            '...i,...j,ijk->...k',
            m_2_Ema, m_0_Ema, EPS)
        M_rEma = np.array([m_0_Ema, m_1_Ema, m_2_Ema])
        M_Emra = np.einsum('rEma->Emra', M_rEma)
        norm_M_Emra = np.sqrt(
            np.einsum('...ij,...ij->...i', M_Emra, M_Emra)
        )[..., np.newaxis]

        T_Emra = M_Emra / norm_M_Emra
        return T_Emra

    B_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B_Eimabc(self):
        # return self.B1_Eimabc
        return np.einsum(
            'Emra,Emsb,Eimabc->Eimrsc',
            self.T_Emra[..., :, :2], self.T_Emra[..., :, :2], self.B1_Eimabc
        )



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch

from traits.api import \
    Property, cached_property, Instance

from ibvpy.tmodel.mats3D.mats3D_cmdm.mats3D_cmdm import \
    MATS3DMicroplaneDamage

from ibvpy.tmodel import \
    MATS2DMicroplaneDamage, PhiFnStrainSoftening, PhiFnGeneral, \
    PhiFnStrainHardening, PhiFnStrainHardeningLinear

# from mathkit.mfn.mfn_polar.mfn_polar import MFnPolar

import numpy as np

class MATS2D5MicroplaneDamage(MATS3DMicroplaneDamage):
    '''
    Degenerated 3D microplane model
        assume linear elastic behavior in eps_z
        assume anistoropic damage in the in-plane components.
        
    needs the strain tensor in the surface coordinates 
    of a shell/plate. 
    '''

    # Specify the class to use for directional dependence
    # mfn_class = MFnPolar

    mats2D_cmdm = Instance(MATS2DMicroplaneDamage)
    def _mats2D_cmdm_default(self):
        return MATS2DMicroplaneDamage()

    # @todo: handle the number of microplanes in a better way
    # currently, the get_n_mp does not return the same value
    # as set_n_mp
    #
    # introduce a hidden parameter in the MATSXD implementation
    # for 3D no user-definable n_mp is avialable - the hidden
    # value of MATSXD is just filled with constant value
    #
    # For MATS2D the public n_mp trait gets defined that 
    # seths the hidden _n_mp
    # For MATS2D5 the public n_mp increased by one is set as
    # the hidden _n_mp 
    #
    n_mp = Property(depends_on = 'mats2D_cmdm.n_mp')
    @cached_property
    def _get_n_mp(self):
        return self.mats2D_cmdm.n_mp + 1
    def _set_n_mp(self, value):
        self.mats2D_cmdm.n_mp + value

    #---------------------------------------------------------------------------------
    # Augmented polar discretization with a single z-normal
    #---------------------------------------------------------------------------------
    # get the normal vectors of the microplanes
    _MPN = Property(depends_on = 'n_mp')
    @cached_property
    def _get__MPN(self):
        mpn2D_arr = self.mats2D_cmdm._MPN
        mpnZ_arr = np.zeros((mpn2D_arr.shape[0], 1), dtype = 'float64')
        mpn3D_arr = np.append(mpn2D_arr, mpnZ_arr, axis = 1)
        return np.append(mpn3D_arr, [[0.0, 0.0, 1.0]], axis = 0)

    # get the weights of the microplanes
    _MPW = Property(depends_on = 'n_mp')
    @cached_property
    def _get__MPW(self):
        mpw2D_arr = self.mats2D_cmdm._MPW
        return np.append(mpw2D_arr, 1.0)

    #-----------------------------------------------------------------------------------------------
    # Get the damage state for all microplanes
    #-----------------------------------------------------------------------------------------------
    def get_phi_arr(self, sctx, e_max_arr):
        '''
        Return the damage coefficients
        '''
        # gather the coefficients for parameters depending on the orientation
        carr_list = [self.varpars[key].polar_fn_vectorized(self.alpha_list)
                     for key in self.phi_fn.identify_parameters() ]
        # vectorize the damage function evaluation
        n_arr = 1 + len(carr_list)
        phi_fn_vectorized = np.frompyfunc(self.phi_fn.get_value, n_arr, 1)
        # damage parameter for each microplane
        phi_arr = np.zeros_like(e_max_arr)
        carr_list = [ carr[:-1] for  carr in carr_list ]
        phi_arr[:-1] = phi_fn_vectorized(e_max_arr[:-1], *carr_list)
        phi_arr[-1] = 1.0
        return phi_arr




'''
'''

import traits.api as tr

from .i_tloop import ITLoop
from .i_tstep import ITStep
from .tline_mixin import TLineMixIn


@tr.provides(ITLoop)
class TLoop(TLineMixIn):
    '''The time loop serves as the base class for application time loops.
    That can be interrupted paused, resumed or restarted.

    The implementation of the loop must contain the break criterion

    while True:
        if self.restart or self.paused:
            break    
        #calculation

    '''

    tstep = tr.WeakRef(ITStep)

    sim = tr.Property

    def _get_sim(self):
        return self.tstep.sim

    hist = tr.Property

    def _get_hist(self):
        return self.tstep.hist

    interrupt = tr.Bool(False)

    def reset(self):
        self.restart = True

    restart = tr.Bool(True)

    user_wants_abort = tr.Property

    def _get_user_wants_abort(self):
        return self.interrupt

    def init(self):
        if self.interrupt:
            self.interrupt = False
        if self.restart:
            self.tline.val = self.tline.min
            self.tstep.init_state()
            self.hist.init_state()
            self.restart = False

    def eval(self):
        '''This method is called by the tloop_thread.
        '''
        raise NotImplementedError

    def __call__(self):
        self.init()
        return self.eval()



'''
'''
from math import sqrt as scalar_sqrt

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from ibvpy.tmodel.mats_eval import IMATSEval
from numpy import \
     array, zeros, dot, float64
from traits.api import \
     Array, Enum, \
     Trait, Event, provides, \
     Dict, Property, cached_property
import bmcs_utils.api as bu

#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
@provides( IMATSEval )
class MATS2D5Bond(MATS2DEval):
    '''
    Elastic Model.
    '''


    #---------------------------------------------------------------------------
    # Parameters of the numerical algorithm (integration)
    #---------------------------------------------------------------------------

    stress_state = Enum("plane_stress", "plane_strain")

    #---------------------------------------------------------------------------
    # Material parameters 
    #---------------------------------------------------------------------------

    E_m = bu.Float(1.,  # 34e+3,
                 label="E_m",
                 desc="Young's Modulus",
                 auto_set=False)
    nu_m = bu.Float(0.2,
                 label='nu_m',
                 desc="Poison's ratio",
                 auto_set=False)

    E_f = bu.Float(1.,  # 34e+3,
                 label="E_f",
                 desc="Young's Modulus",
                 auto_set=False)
    nu_f = bu.Float(0.2,
                 label='nu_f',
                 desc="Poison's ratio",
                 auto_set=False)

    G = bu.Float(1.,  # 34e+3,
                 label="G",
                 desc="Shear Modulus",
                 auto_set=False)

    D_el = Property(Array(float), depends_on='E_f, nu_f,E_m,nu_f,G, stress_state')

    @cached_property
    def _get_D_el(self):
        if self.stress_state == "plane_stress":
            return self._get_D_plane_stress()
        else:
            return self._get_D_plane_strain()

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #---------------------------------------------------------------------------------------------
    # View specification
    #---------------------------------------------------------------------------------------------

    view_traits = bu.View(
        bu.Item('E_m'),
        bu.Item('nu_m'),
        bu.Item('E_f'),
        bu.Item('nu_f'),
        bu.Item('G')
    )

    def get_corr_pred(self, eps_app_eng, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''

        sigma = dot(self.D_el[:], eps_app_eng)

        # You print the stress you just computed and the value of the apparent E

        return  sigma, self.D_el

    #---------------------------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #---------------------------------------------------------------------------------------------

    def _get_D_plane_stress(self):
        E_m = self.E_m
        nu_m = self.nu_m
        E_f = self.E_f
        nu_f = self.nu_f
        G = self.G
        D_stress = zeros([8, 8])
        D_stress[0, 0] = E_m / (1.0 - nu_m * nu_m)
        D_stress[0, 1] = E_m / (1.0 - nu_m * nu_m) * nu_m
        D_stress[1, 0] = E_m / (1.0 - nu_m * nu_m) * nu_m
        D_stress[1, 1] = E_m / (1.0 - nu_m * nu_m)
        D_stress[2, 2] = E_m / (1.0 - nu_m * nu_m) * (1.0 / 2.0 - nu_m / 2.0)

        D_stress[3, 3] = E_f / (1.0 - nu_f * nu_f)
        D_stress[3, 4] = E_f / (1.0 - nu_f * nu_f) * nu_f
        D_stress[4, 3] = E_f / (1.0 - nu_f * nu_f) * nu_f
        D_stress[4, 4] = E_f / (1.0 - nu_f * nu_f)
        D_stress[5, 5] = E_f / (1.0 - nu_f * nu_f) * (1.0 / 2.0 - nu_f / 2.0)

        D_stress[6, 6] = G
        D_stress[7, 7] = G
        return D_stress

    def _get_D_plane_strain(self):
        # TODO: adapt to use arbitrary 2d model following the 1d5 bond
        E_m = self.E_m
        nu_m = self.nu_m
        E_f = self.E_f
        nu_f = self.nu_f
        G = self.G
        D_strain = zeros([8, 8])
        D_strain[0, 0] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (1.0 - 2.0 * nu_m)
        D_strain[0, 1] = E_m / (1.0 + nu_m) / (1.0 - 2.0 * nu_m) * nu_m
        D_strain[1, 0] = E_m / (1.0 + nu_m) / (1.0 - 2.0 * nu_m) * nu_m
        D_strain[1, 1] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (1.0 - 2.0 * nu_m)
        D_strain[2, 2] = E_m * (1.0 - nu_m) / (1.0 + nu_m) / (2.0 - 2.0 * nu_m)

        D_strain[3, 3] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (1.0 - 2.0 * nu_f)
        D_strain[3, 4] = E_f / (1.0 + nu_f) / (1.0 - 2.0 * nu_f) * nu_f
        D_strain[4, 3] = E_f / (1.0 + nu_f) / (1.0 - 2.0 * nu_f) * nu_f
        D_strain[4, 4] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (1.0 - 2.0 * nu_f)
        D_strain[5, 5] = E_f * (1.0 - nu_f) / (1.0 + nu_f) / (2.0 - 2.0 * nu_f)

        D_strain[6, 6] = G
        D_strain[7, 7] = G
        return D_strain

    #---------------------------------------------------------------------------------------------
    # Response trace evaluators
    #---------------------------------------------------------------------------------------------

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([ scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2) ])

    def get_eps_app_m(self, sctx, eps_app_eng):
        return self.map_eps_eng_to_mtx((eps_app_eng[:3]))

    def get_eps_app_f(self, sctx, eps_app_eng):
        return self.map_eps_eng_to_mtx((eps_app_eng[3:6]))

    def get_sig_app_m(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return self.map_sig_eng_to_mtx((sig_eng[:3]))

    def get_sig_app_f(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return self.map_sig_eng_to_mtx((sig_eng[3:6]))

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {
                'eps_app_f'  : self.get_eps_app_f,
                'eps_app_m'  : self.get_eps_app_m,
                'sig_app_f'  : self.get_sig_app_f,
                'sig_app_m'  : self.get_sig_app_m,
                'sig_norm'   : self.get_sig_norm}




'''
Created on 08.02.2018

@author: abaktheer

Microplane damage model 2D - ODFS (Wu [2009])
'''

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from numpy import \
    array, einsum, identity, sqrt
from traits.api import \
    Constant, \
    Float, Property, cached_property

from ibvpy.tmodel.matsXD.vmatsXD_eval import MATSXDEval
import numpy as np
import traits.api as tr


class MATSXDMicroplaneDamageODF(MATSXDEval):

    epsilon_0 = Float(59e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    epsilon_f = Float(250e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    c_T = Float(0.00,
                label="a",
                desc="Lateral pressure coefficient",
                enter_set=True,
                auto_set=False)

    zeta_G = Float(1.0,
                   label="zeta_G",
                   desc="anisotropy parameter",
                   enter_set=True,
                   auto_set=False)

    state_var_shapes = tr.Property(tr.Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''

    @cached_property
    def _get_state_var_shapes(self):
        return {'kappa': (self.n_mp,),
                'omega': (self.n_mp,)}

    #-------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    #-------------------------------------------------------------------------

    # get the dyadic product of the microplane normals
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        # dyadic product of the microplane normals

        MPNN_nij = einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')

    @cached_property
    def _get__MPTT(self):
        # Third order tangential tensor for each microplane
        delta = identity(2)
        MPTT_nijr = 0.5 * (einsum('ni,jr -> nijr', self._MPN, delta) +
                           einsum('nj,ir -> njir', self._MPN, delta) - 2.0 *
                           einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN))
        return MPTT_nijr

    def _get_e_Emna(self, eps_Emab):
        # Projection of apparent strain onto the individual microplanes
        e_ni = einsum('nb,Emba->Emna', self._MPN, eps_Emab)
        return e_ni

    def _get_e_N_Emn(self, e_Emna):
        # get the normal strain array for each microplane
        e_N_Emn = einsum('Emna, na->Emn', e_Emna, self._MPN)
        return e_N_Emn

    def _get_e_equiv_Emn(self, e_Emna):
        '''
        Returns a list of the microplane equivalent strains
        based on the list of microplane strain vectors
        '''
        # magnitude of the normal strain vector for each microplane
        e_N_Emn = self._get_e_N_Emn(e_Emna)
        # positive part of the normal strain magnitude for each microplane
        e_N_pos_Emn = (np.abs(e_N_Emn) + e_N_Emn) / 2
        # normal strain vector for each microplane
        e_N_Emna = einsum('Emn,ni -> Emni', e_N_Emn, self._MPN)
        # tangent strain ratio
        c_T = self.c_T
        # tangential strain vector for each microplane
        e_T_Emna = e_Emna - e_N_Emna
        # squared tangential strain vector for each microplane
        e_TT_Emn = einsum('Emni,Emni -> Emn', e_T_Emna, e_T_Emna)
        # equivalent strain for each microplane
        e_equiv_Emn = sqrt(e_N_pos_Emn * e_N_pos_Emn + c_T * e_TT_Emn)
        return e_equiv_Emn

    def update_state_variables(self, eps_Emab, kappa, omega):

        e_Emna = self._get_e_Emna(eps_Emab)
        eps_eq_Emn = self._get_e_equiv_Emn(e_Emna)
        f_trial_Emn = eps_eq_Emn - self.epsilon_0
        I = np.where(f_trial_Emn > 0)
        kappa[I] = eps_eq_Emn[I]
        omega[I] = self._get_omega(eps_eq_Emn[I])
        return I

    def _get_omega(self, kappa_Emn):
        '''
        Return new value of damage parameter
        @param kappa:
        '''
        omega_Emn = np.zeros_like(kappa_Emn)
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        kappa_idx = np.where(kappa_Emn >= epsilon_0)
        omega_Emn[kappa_idx] = (
            1.0 - (epsilon_0 / kappa_Emn[kappa_idx] *
                   np.exp(-1.0 * (kappa_Emn[kappa_idx] - epsilon_0) /
                          (epsilon_f - epsilon_0))
                   ))
        return omega_Emn

    def _get_phi_Emab(self, kappa_Emn):
        # Returns the 2nd order damage tensor 'phi_mtx'
        # scalar integrity factor for each microplane
        phi_Emn = 1.0 - self._get_omega(kappa_Emn)
        # integration terms for each microplanes
        phi_Emab = einsum('Emn,n,nab->Emab', phi_Emn, self._MPW, self._MPNN)
        return phi_Emab

#     def _get_beta_Emabcd(self, phi_Emab):
#         '''
#         Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
#         (cf. [Jir99], Eq.(21))
#         '''
#         delta = identity(2)
#         beta_Emijkl = 0.25 * (einsum('Emik,jl->Emijkl', phi_Emab, delta) +
#                               einsum('Emil,jk->Emijkl', phi_Emab, delta) +
#                               einsum('Emjk,il->Emijkl', phi_Emab, delta) +
#                               einsum('Emjl,ik->Emijkl', phi_Emab, delta))
#
#         return beta_Emijkl

    #----------------------------------------------------------------
    #  the fourth order volumetric-identity tensor
    #----------------------------------------------------------------
    def _get_I_vol_abcd(self):

        delta = identity(2)
        I_vol_abcd = (1.0 / 3.0) * einsum('ab,cd -> abcd', delta, delta)
        return I_vol_abcd

    #----------------------------------------------------------------
    # Returns the fourth order deviatoric-identity tensor
    #----------------------------------------------------------------
    def _get_I_dev_abcd(self):

        delta = identity(2)
        I_dev_abcd = 0.5 * (einsum('ac,bd -> abcd', delta, delta) +
                            einsum('ad,bc -> abcd', delta, delta)) \
            - (1. / 3.0) * einsum('ab,cd -> abcd', delta, delta)

        return I_dev_abcd

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_vol_ab(self):

        delta = identity(2)
        P_vol_ab = (1.0 / 3.0) * delta
        return P_vol_ab

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_dev [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_dev_nabc(self):

        delta = identity(2)
        P_dev_nabc = 0.5 * einsum('nd,da,bc -> nabc', self._MPN, delta, delta)
        return P_dev_nabc

    #----------------------------------------------------------------
    # Returns the outer product of P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_PP_vol_abcd(self):

        delta = identity(2)
        PP_vol_abcd = (1.0 / 9.0) * einsum('ab,cd -> abcd', delta, delta)
        return PP_vol_abcd

    #----------------------------------------------------------------
    # Returns the inner product of P_dev
    #----------------------------------------------------------------
    def _get_PP_dev_nabcd(self):

        delta = identity(2)
        PP_dev_nabcd = 0.5 * (0.5 * (einsum('na,nc,bd -> nabcd', self._MPN, self._MPN, delta) +
                                     einsum('na,nd,bc -> nabcd', self._MPN, self._MPN, delta)) +
                              0.5 * (einsum('ac,nb,nd -> nabcd', delta, self._MPN, self._MPN) +
                                     einsum('ad,nb,nc -> nabcd', delta, self._MPN, self._MPN))) - \
            (1.0 / 3.0) * (einsum('na,nb,cd -> nabcd', self._MPN, self._MPN, delta) +
                           einsum('ab,nc,nd -> nabcd', delta, self._MPN, self._MPN)) + \
            (1.0 / 9.0) * einsum('ab,cd -> abcd', delta, delta)

        return PP_dev_nabcd

    #--------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(29))
    #--------------------------------------------------------------------------
    def _get_S_1_Emabcd(self, eps_Emab, kappa, omega):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        phi_Emn = 1.0 - self._get_omega(kappa)

        PP_vol_abcd = self._get_PP_vol_abcd()
        PP_dev_nabcd = self._get_PP_dev_nabcd()
        I_dev_abcd = self._get_I_dev_abcd()

        S_1_Emabcd = K0 * einsum('Emn, n, abcd-> Emabcd', phi_Emn, self._MPW, PP_vol_abcd) + \
            G0 * 2.0 * self.zeta_G * einsum('Emn, n, nabcd-> Emabcd',
                                            phi_Emn, self._MPW, PP_dev_nabcd) - (1.0 / 3.0) * (
                2.0 * self.zeta_G - 1.0) * G0 * einsum('Emn, n, abcd-> Emabcd',
                                                       phi_Emn, self._MPW, I_dev_abcd)

        return S_1_Emabcd

#     #------------------------------------------
#     # scalar damage factor for each microplane
#     #------------------------------------------
#     def _get_d_Em(self, s_Emng, eps_Emab):
#
#         d_Emn = 1.0 - self.get_state_variables(s_Emng, eps_Emab)[0]
#
#         d_Em = (1.0 / 3.0) * einsum('Emn,n-> Em',  d_Emn, self._MPW)
#
#         return d_Em
#
#     #------------------------------------------
#     # The 4th order volumetric damage tensor
#     #------------------------------------------
#     def _get_M_vol_abcd(self, sctx, eps_app_eng, sigma_kk):
#
#         d = self._get_Em( s_Emng, eps_Emab)
#         delta = identity(2)
#
#         I_4th_abcd = 0.5 * (einsum('ac,bd -> ijkl', delta, delta) +
#                             einsum('il,jk -> ijkl', delta, delta))
#
#         # print 'M_vol', (1 - d) * I_4th_ijkl
#
#         return (1 - d) * I_4th_ijkl
#
#     #------------------------------------------
#     # The 4th order deviatoric damage tensor
#     #------------------------------------------
#     def _get_M_dev_tns(self, phi_mtx):
#
#         delta = identity(3)
#         I_4th_ijkl = 0.5 * (einsum('ik,jl -> ijkl', delta, delta) +
#                             einsum('il,jk -> ijkl', delta, delta))
#         tr_phi_mtx = np.trace(phi_mtx)
#
#         M_dev_ijkl = self.zeta_G * (0.5 * (einsum('ik,jl->ijkl', delta, phi_mtx) +
#                                            einsum('il,jk->ijkl', delta, phi_mtx)) +
#                                     0.5 * (einsum('ik,jl->ijkl', phi_mtx, delta) +
#                                            einsum('il,jk->ijkl', phi_mtx, delta))) \
#             - (2. * self.zeta_G - 1.) * (tr_phi_mtx / 3.) * I_4th_ijkl
#
#         return M_dev_ijkl
#
#     #--------------------------------------------------------------------------
#     # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(31))
#     #--------------------------------------------------------------------------
#     def _get_S_2_Emabcd(self, sctx, eps_app_eng, sigma_kk):
#
#         K0 = self.E / (1. - 2. * self.nu)
#         G0 = self.E / (1. + self.nu)
#
#         I_vol_ijkl = self._get_I_vol_4()
#         I_dev_ijkl = self._get_I_dev_4()
#         phi_mtx = self._get_phi_mtx(sctx, eps_app_eng, sigma_kk)
#         M_vol_ijkl = self._get_M_vol_tns(sctx, eps_app_eng, sigma_kk)
#         M_dev_ijkl = self._get_M_dev_tns(phi_mtx)
#
#         S_2_ijkl = K0 * einsum('ijmn,mnrs,rskl -> ijkl', I_vol_ijkl, M_vol_ijkl, I_vol_ijkl ) \
#             + G0 * einsum('ijmn,mnrs,rskl -> ijkl', I_dev_ijkl, M_dev_ijkl, I_dev_ijkl)\
#
#         return S_2_ijkl
#
#     #--------------------------------------------------------------------------
#     # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(34))
#     #--------------------------------------------------------------------------
#     def _get_S_3_Emabcd(self, sctx, eps_app_eng, sigma_kk):
#
#         K0 = self.E / (1. - 2. * self.nu)
#         G0 = self.E / (1. + self.nu)
#
#         I_vol_ijkl = self._get_I_vol_4()
#         I_dev_ijkl = self._get_I_dev_4()
#
#         # The fourth order elastic stiffness tensor
#         S_0_ijkl = K0 * I_vol_ijkl + G0 * I_dev_ijkl
#
#         d_n = self._get_state_variables(sctx, eps_app_eng, sigma_kk)[:, 5]
#
#         PP_vol_4 = self._get_PP_vol_4()
#         PP_dev_4 = self._get_PP_dev_4()
#
#         delta = identity(3)
#         I_4th_ijkl = einsum('ik,jl -> ijkl', delta, delta)
#
#         D_ijkl = einsum('n,n,ijkl->ijkl', d_n, self._MPW, PP_vol_4) + \
#             2 * self.zeta_G * einsum('n,n,nijkl->ijkl', d_n, self._MPW, PP_dev_4) - (
#                 1 / 3.) * (2 * self.zeta_G - 1) * einsum('n,n,ijkl->ijkl', d_n, self._MPW, I_dev_ijkl)
#
#         phi_ijkl = (I_4th_ijkl - D_ijkl)
#
#         S_ijkl = einsum('ijmn,mnkl', phi_ijkl, S_0_ijkl)
#
#         return S_ijkl
#
    #-------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor using (double orthotropic) assumption
    #-------------------------------------------------------------------------
    def _get_S_4_Emabcd(self, eps_Emab, kappa, omega):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        I_vol_abcd = self._get_I_vol_abcd()
        I_dev_abcd = self._get_I_dev_abcd()
        delta = identity(2)

        phi_Emab = self._get_phi_Emab(kappa)
        D_Emab = delta - phi_Emab
        d_Em = (1.0 / 3.0) * np.trace(D_Emab)
        D_bar_Emab = self.zeta_G * (D_Emab - d_Em * delta)

        S_4_Emabcd = (1.0 - d_Em) * K0 * I_vol_abcd + (1.0 - d_Em) * G0 * I_dev_abcd + (2.0 / 3.0) * (G0 - K0) * \
            (einsum('ij,Emkl -> Emijkl', delta, D_bar_Emab) +
             einsum('Emij,kl -> Emijkl', D_bar_Emab, delta)) + 0.5 * (-K0 + 2.0 * G0) * \
            (0.5 * (einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab) + einsum('Emil,jk -> Emijkl', D_bar_Emab, delta)) +
             0.5 * (einsum('Emil,jk -> Emijkl', D_bar_Emab, delta) + einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab)))

        return S_4_Emabcd

    def _get_S_5_Emabcd(self, eps_Emab, kappa, omega):
        #----------------------------------------------------------------------
        # Returns the fourth order secant stiffness tensor (restrctive orthotropic)
        #----------------------------------------------------------------------

        K0 = self.E / (1. - 2. * self.nu)
        G0 = self.E / (1. + self.nu)

        delta = identity(2)
        phi_Emab = self._get_phi_Emab(kappa)
        # damaged stiffness without simplification
        S_5_Emabcd = (1.0 / 3.0) * (K0 - G0) * 0.5 * ((einsum('ij,Emkl -> Emijkl', delta, phi_Emab) +
                                                       einsum('Emij,kl -> Emijkl', phi_Emab, delta))) + \
            G0 * 0.5 * ((0.5 * (einsum('ik,Emjl -> Emijkl', delta, phi_Emab) + einsum('Emil,jk -> Emijkl', phi_Emab, delta)) +
                         0.5 * (einsum('Emik,jl -> ijkl', phi_Emab, delta) + einsum('il,Emjk  -> Emijkl', delta, phi_Emab))))

        # print 'S_5_Emabcd', S_5_Emabcd
        return S_5_Emabcd
#
#     #-------------------------------------------------------------------------
#     # Returns the fourth order secant stiffness tensor (double orthotropic N-T split)
#     #-------------------------------------------------------------------------
#     def _get_S_5_Emabcd(self, sctx, eps_app_eng, sigma_kk):
#
#         E_N = self.E / (3.0 - 2.0 * (1.0 + self.nu))
#         E_T = self.E / (1. + self.nu)
#
#         I_vol_ijkl = self._get_I_vol_4()
#         I_dev_ijkl = self._get_I_dev_4()
#         delta = identity(3)
#         phi_mtx = self._get_phi_mtx(sctx, eps_app_eng, sigma_kk)
#         D_ij = delta - phi_mtx
#         d = (1. / 3.) * np.trace(D_ij)
#         D_bar_ij = self.zeta_G * (D_ij - d * delta)
#
#         S_5_ijkl = (1 - d) * E_N * I_vol_ijkl + (1 - d) * E_T * I_dev_ijkl + (2 / 3.) * (E_T - E_N) * \
#             (einsum('ij,kl -> ijkl', delta, D_bar_ij) +
#              einsum('ij,kl -> ijkl', D_bar_ij, delta)) + 0.5 * (2 * E_T - E_N) *\
#             (0.5 * (einsum('ik,jl -> ijkl', delta, D_bar_ij) + einsum('il,jk -> ijkl', D_bar_ij, delta)) +
#              0.5 * (einsum('il,jk -> ijkl', D_bar_ij, delta) + einsum('ik,jl -> ijkl', delta, D_bar_ij)))
#
#         return S_5_ijkl

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab_n1, tn1, kappa, omega):

        I = self.update_variables(eps_Emab_n1, kappa, omega)

        #----------------------------------------------------------------------
        # if the regularization using the crack-band concept is on calculate the
        # effective element length in the direction of principle strains
        #----------------------------------------------------------------------
        # if self.regularization:
        #    h = self.get_regularizing_length(sctx, eps_app_eng)
        #    self.phi_fn.h = h

#         #------------------------------------------------------------------
#         # Damage tensor (2th order):
#         #------------------------------------------------------------------
#         phi_Emab = self._get_phi_Emab(kappa_Emn)
#
#         #------------------------------------------------------------------
#         # Damage tensor (4th order) using product- or sum-type symmetrization:
#         #------------------------------------------------------------------
#         beta_Emabcd = self._get_beta_Emabcd(phi_Emab)
#
#         #------------------------------------------------------------------
#         # Damaged stiffness tensor calculated based on the damage tensor beta4:
#         #------------------------------------------------------------------
#         D_Emabcd, = einsum(
#             'Emijab, abef, Emabef -> Emijab', beta_Emabcd, self.D_abef, beta_Emabcd)

        D_Emabcd = self._get_S_1_Emabcd(eps_Emab_n1, kappa, omega)

        sig_Emab = einsum('Emabcd,Emcd -> Emab', D_Emabcd, eps_Emab_n1)

        return D_Emabcd, sig_Emab


class MATS2DMplDamageODF(MATSXDMicroplaneDamageODF, MATS2DEval):

    # implements(IMATSEval)

    #-----------------------------------------------
    # number of microplanes - currently fixed for 3D
    #-----------------------------------------------
    n_mp = Constant(28)

    _alpha_list = Property(depends_on='n_mp')

    @cached_property
    def _get__alpha_list(self):
        return array([np.pi / self.n_mp * (i - 0.5)
                      for i in range(1, self.n_mp + 1)])

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return array([[np.cos(alpha), np.sin(alpha)] for alpha in self._alpha_list])

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return np.ones(self.n_mp) / self.n_mp * 2.0



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 18, 2009 by: rch

from traits.api import HasTraits, Instance

import unittest

from ibvpy.fets.fets2D.fets2D4q import \
    FETS2D4Q
from ibvpy.util.simgrid import \
    simgrid
from ibvpy.tmodel.mats_eval import MATSEval

class TestMATS2D(unittest.TestCase, HasTraits):

    mats_eval = Instance( MATSEval )
    
    def setUp(self):

        self.fets_eval  = FETS2D4Q( mats_eval  = self.mats_eval )

        self.support_slices = [
                          [ (0   ,slice(None),0   ,slice(None)), # y axis 0
                            (0   ,0          ,0   ,0          ), # (0,0)   1
                          ],
                          [ 
                            (0   ,0   ,0   ,0      ), #  origin   0
                            (slice(None),0   ,slice(None),0   ), # xz plane  1
                          ],
                          ]
        self.support_dirs = [[0],[1]]
        
        self.loading_slices = [ 
                          (-1  ,slice(None),-1  ,slice(None)),  # loading in x dir
                          (slice(None),-1  ,slice(None),-1  ),  # loading in y dir
                        ]

    def assert_symmetry(self, load_dirs = [0,1] ):
        '''Assert that the symmetry is given for the applied loadings.
        '''
        load = 0.0001

        r = [ simgrid( self.fets_eval,  (1,1), (1,1),
                       support_slice, self.support_dirs,
                       loading_slice, load_dir, 
                       load, 1, vars = [] )
              for support_slice, loading_slice, load_dir 
              in zip( self.support_slices, self.loading_slices, load_dirs ) ]
        
        self.assertAlmostEqual( r[0][1][-1], r[1][1][-2] )

    def assert_stress_value(self,
                            sig_expected,  
                            n_steps = 3, 
                            load = 0.0001,
                            ):
        '''Assert that the symmetry is given for the applied loadings.
        '''
        vars = ['sig_app']

        r = simgrid( self.fets_eval,  (1,1), (1,1),
                       self.support_slices[0], self.support_dirs,
                       self.loading_slices[0], 0, 
                       load, n_steps, vars )

        sig_end = r[2][0]
        for sig, sig_exp in zip( sig_end.flatten(), sig_expected ):
            self.assertAlmostEqual( sig, sig_exp, 5 )

    def assert_total_energy_value(self,
                                         value_expected,  
                                         ivar = 'strain_energy',
                                         n_steps = 3, 
                                         load = 1.0,
                            ):
        '''Assert that the symmetry is given for the applied loadings.
        '''
        ivars = [ivar]
        r = simgrid( self.fets_eval,  (1,1), (1,1),
                     self.support_slices[0], self.support_dirs,
                     self.loading_slices[0], 0, 
                     load, n_steps, ivars = ivars )

        self.assertAlmostEqual( r[3][0][0], value_expected )
        
        


'''
Created on Dec 17, 2018

@author: rch
'''

from ibvpy.view.ui.bmcs_tree_node import \
    BMCSTreeNode

import traits.api as tr

from .i_tmodel import ITModel
from .tloop import TLoop
from .tstep import TStep


@tr.provides(ITModel)
class TModel(BMCSTreeNode):
    '''
    Base class of a time-stepping model capturing the state
    of a modeled object at the particular instance of time 
    supplying the predictor and corrector interface.

    The model includes the geometry, boundary conditions, 
    material characteristics, type of material behavior

    A model communicates with the time loop through their 
    common interface which depends on the particular type 
    of time-stepping scheme. In case of a time integration 
    scheme the control variable is represented by displacements,
    in case of an optimization problem, the function supplying 
    the values and gradients of response variables is required.
    In case of fragmentation problem, the model supplies 
    the array of load factors need to induce a crack in 
    a material point. 
    '''

    tloop_type = tr.Type(TLoop)
    '''Type of time loop to be used with the model
    '''

    tstep_type = tr.Type(TStep)
    '''Type of the time step to be used with the model
    '''

    def init_state(self):
        pass

    def record_state(self):
        pass



'''
Created on Jul 9, 2017

@author: rch
'''

from io import StringIO
from traits.api import \
    HasStrictTraits, Interface, Str, \
    Dict, Property, provides, Missing
import numpy as np


class IRItem(Interface):
    '''Interface supporting the protocol or a report item. 
    '''

    def write_report_entry(self):
        pass


class RItem(HasStrictTraits):
    '''Report item.
    '''


@provides(IRItem)
class RInputRecord(RItem):
    '''This class scans the parameters
    that have as metadata one of the input tags and 
    specified in the input tag list.
    '''

    name = Str

    itags = Dict

    def get_subrecords(self):
        subrecords = {}
        for name, value in list(self.trait_get(report=True).items()):
            subrecords[name] = value
        return subrecords

    def get_subrecord_traits(self):
        return self.traits(report=True)

    def _get_rinput_traits(self, itags=None):
        if itags == None:
            itags = self.itags
        input_traits = {}
        for itag in list(itags.keys()):
            input_traits.update(self.traits(**{itag: True}))
        return input_traits

    def yield_rinput_traits(self, path='', itags=None):
        return {path + '.' + name: trait
                for name, trait in list(self._get_rinput_traits(itags).items())
                }

    def _get_rinputs(self, itags=None):
        if itags == None:
            itags = self.itags
        rinput_traits = {}
        for itag in list(itags.keys()):
            rinput_traits.update(self.trait_get(**{itag: True}))
        return rinput_traits

    def _repr_latex_(self):
        io = StringIO()
        io.write(r'''
        \begin{array}{lrrl}\hline
        ''')
        itraits = self._get_rinput_traits()
        for name, trait in list(itraits.items()):
            value = getattr(self, name, Missing)
            unit = trait.unit
            symbol = trait.symbol
            label = trait.label
            desc = trait.desc
            if symbol:
                io.write(r'''\textrm{%s} & %s = %s & \textrm{[%s]} & \textrm{%s}  \\
                ''' % (name, symbol, str(value), unit, desc))
            else:
                io.write(r'''\textrm{%s} & \textrm{%s} & & \textrm{%s}  \\
                ''' % (name, str(value), desc))
        io.write(r'''\hline
        ''')
        records = self.get_subrecord_traits()
        for name, trait in list(records.items()):
            value = getattr(self, name, Missing)
            clname = value.__class__.__name__
            desc = trait.desc
            io.write(r'''\textbf{%s} & \textrm{%s} & & \textrm{%s} \\
            ''' % (name, clname, desc))
        io.write(r'''\hline
        \end{array}
        ''')
        return io.getvalue()

    def write_tex_table_record_entries(self, f, path='', itags=None):
        itraits = self._get_rinput_traits(itags)
        for name, trait in list(itraits.items()):
            value = getattr(self, name, Missing)
            unit = trait.unit
            symbol = trait.symbol
            label = trait.label
            desc = trait.desc
            f.write(r'''\texttt{%s%s} & $%s$ = %s [%s] & {\footnotesize %s}  \\
            ''' %
                    (path, name.replace('_', '\_'), symbol, str(value), unit, desc))

    def write_tex_table_record(self, f, path='', rdir=None,
                               rel_study_path=None,
                               itags=None):
        tag_io = StringIO()
        self.write_tex_table_record_entries_with_figs(
            tag_io, path, rdir, rel_study_path, itags)
        f.write(tag_io.getvalue())
        records = self.get_subrecords()
        for name, record in list(records.items()):
            tag_io = StringIO()
            clname = record.__class__.__name__
            nm = name.replace('_', '\_')
            record.write_tex_table_record(
                tag_io, path + nm + '.', rdir, rel_study_path, itags)
            tag_string = tag_io.getvalue()
            if len(tag_string) > 0:
                f.write(r'''\midrule
\multicolumn{3}{l}{\textbf{\textsf{%s: %s}}}\\

''' % (clname, name.replace('_', '\_')))
                f.write(tag_string)

    def write_tex_table_record_entries_with_figs(self, f, path='', rdir=None,
                                                 rel_study_path=None,
                                                 itags=None):
        entries_io = StringIO()
        figures_io = StringIO()
        self.write_tex_table_record_entries(entries_io, path, itags)
        if rdir != None:
            self.write_figure(figures_io, rdir, rel_study_path)
        entries_str = entries_io.getvalue()
        figures_str = figures_io.getvalue()
        f.write(entries_str)
        f.write(figures_str)

    def write_figure(self, f, rdir=None, rel_study_path=None):
        pass

    inputs = Property()

    def _get_inputs(self):
        inputs = StringIO()
        self.write_tex_table(inputs)
        return inputs.getvalue()

    def write_tex_table(self, f, rdir=None, rel_study_path=None, itags=None):
        f.write(r'''
{\scriptsize 
\begin{longtable}{lrp{4cm}}\toprule
\textbf{\textsf{Model parameter}} 
& 
\textbf{\textsf{Symbol = Value [Unit]}} 
&
\textbf{\textsf{Description}}  \\\midrule \midrule
''')
        self.write_tex_table_record(f, '', rdir, rel_study_path, itags)
        f.write(r'''\bottomrule 
\end{longtable}
}
''')

    def yield_rinput_section(self, path='', itags=None):
        rinput_traits = self.yield_rinput_traits(path, itags)
        for name, trait in list(self.get_subrecords().items()):
            rinput_traits.update(
                trait.yield_rinput_section(path + name + '.', itags))
        return rinput_traits


class ROutputItem(RItem):

    def write_record(self, f, rdir, rel_study_path=None, itags=None):
        tag_io = StringIO()
        self.write_figure(tag_io, rdir, rel_study_path)
        f.write(tag_io.getvalue())


@provides(IRItem)
class ROutputSection(RItem):

    def write_records(self, f, rdir=None, rel_study_path=None, itags=None):
        records = self.get_subrecords()
        record_fig_list = []
        for record in records:
            tag_io = StringIO()
            record.write_record(tag_io, rdir, rel_study_path, itags)
            tag_string = tag_io.getvalue()
            if len(tag_string) > 0:
                record_fig_list.append(tag_string)
        record_arr = np.array(record_fig_list).reshape(-1, 2)
        for r1, r2 in record_arr:
            f.write(r'''
\noindent
\begin{longtable}{L{7.5cm}L{7.5cm}}''')
            f.write(r'%s & %s \\' % (r1, r2))
            f.write(r'''\end{longtable}
''')

    def write_report(self, f='', rdir=None, itags=None):
        self.write_record(f, rdir, itags)



'''
Created on Jan 23, 2018

@author: rch
'''

import numpy as np

ONE = np.ones((1,), dtype=np.float64)
DELTA = np.identity(3)
DELTA2D = np.identity(2)
DELTA1D = np.identity(1)

# Levi Civita symbol
EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1

EPS2D = np.zeros((2, 2), dtype='f')
EPS2D[(0, 1)] = 1
EPS2D[(1, 0)] = -1

I_sym = (np.einsum('ac,bd->abcd', DELTA, DELTA) +
         np.einsum('ad,bc->abcd', DELTA, DELTA)) / 2.0

# expansion tensor
DELTA23_ab = np.array([[1, 0, 0],
                       [0, 1, 0]], dtype=np.float64)


if __name__ == '__main__':
    print(I_sym)
    print(np.einsum('pq,pq->pq', EPS2D, EPS2D))
    # Convert the tensor to an engineering tensor
    eps_tns = np.array([[[1, 2, 3],
                         [2, 1, 4],
                         [3, 4, 1]]], dtype=np.float64)
    eps_eng1 = np.einsum(
        'ki,...ki->...k', DELTA, eps_tns
    )
    eps_eng2 = 0.5 * np.einsum(
        'kij,...ij->...k', np.fabs(EPS), eps_tns
    )
    eps_eng = np.concatenate((eps_eng1, eps_eng2), axis=-1)
    print('eps_eng', eps_eng)

    dim2 = int(eps_eng.shape[-1] / 2)
    eps = np.einsum(
        'ki,...k->...ki', DELTA, eps_eng[..., :dim2]
    ) + np.einsum(
        'kij,...k->...ij', np.fabs(EPS), eps_eng[..., dim2:]
    )

    print('eps', eps)



'''
Created on Mar 4, 2017

@author: rch
'''

from ibvpy.mathkit.mfn import MFnLineArray
from traits.api import \
    List, Float, Int, Range, Property, \
    cached_property, Bool, Callable, on_trait_change
from ibvpy.view.plot2d import Vis2D, Viz2D
from ibvpy.view.ui.bmcs_tree_node import BMCSLeafNode
import numpy as np


class TFViz2D(Viz2D):

    def plot(self, ax, vot):
        t, y = self.vis2d.get_ty_data(vot)
        ax.plot(t, y)
        y_min, y_max = np.min(y), np.max(y)
        ax.plot([vot, vot], [y_min, y_max])


class TFunPWLInteractive(MFnLineArray, BMCSLeafNode, Vis2D):
    '''Interactive time function.
    '''
    node_name = 'time function'
    t_values = List(Float, [0])
    f_values = List(Float, [0])

    def reset(self):
        self.f_values = [0]
        self.t_values = [0]
        self.f_value = self.f_min

    n_f_values = Int(10,
                     input=True,
                     auto_set=False, enter_set=True)

    f_min = Float(0.0, input=True,
                  auto_set=False, enter_set=True,
                  label='F minimum')

    f_max = Float(1.0,
                  input=True,
                  auto_set=False, enter_set=True,
                  label='F maximum')

    t_ref = Float(1.0, auto_set=False, enter_set=True,
                  label='Initial time range')

    f_value = Range(low='f_min', high='f_max', value=0,
                    input=True,
                    auto_set=False, enter_set=True)

    enable_slider = Bool(True, disable_on_run=True)

    run_eagerly = Bool(True, label='Run eagerly')

    t_snap = Float(0.1, label='Time step to snap to',
                   auto_set=False, enter_set=True)

    def __init__(self, *arg, **kw):
        super(TFunPWLInteractive, self).__init__(*arg, **kw)
        self.xdata = np.array(self.t_values)
        self.ydata = np.array(self.f_values)

    d_t = Property(depends_on='t_ref,n_f_values')

    @cached_property
    def _get_d_t(self):
        return self.t_ref / self.n_f_values

    def _update_xy_arrays(self):
        delta_f = self.f_value - self.f_values[-1]
        self.f_values.append(self.f_value)
        rel_step = delta_f / (self.f_max - self.f_min)
        delta_t = rel_step * self.t_ref
        t_value = np.fabs(delta_t) + self.t_values[-1]
        n_steps = int(t_value / self.t_snap) + 1
        t_value = n_steps * self.t_snap
        self.t_values.append(t_value)
        self.xdata = np.array(self.t_values)
        self.ydata = np.array(self.f_values)
        self.replot()

    def _f_value_changed(self):
        self._update_xy_arrays()
        t_value = self.t_values[-1]
        f_value = self.f_values[-1]
        if self.ui:
            self.ui.model.set_tmax(t_value)
            if self.run_eagerly:
                print('LS-run', t_value, f_value)
                self.ui.run()

    def get_ty_data(self, vot):
        return self.t_values, self.f_values

    viz2d_classes = {
        'time_function': TFViz2D,
    }




from ibvpy.api import RTDofGraph
from numpy import array, ones, trapz, frompyfunc, dot, zeros, fabs


class MATS2DMicroplaneDamageTraceGfmic( RTDofGraph ):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''
#    E_t = zeros((1,))
#    U_t = zeros((1,))
#    G_f = zeros((1,))
    
    def __init__(self,**kwrds):
        super(MATS2DMicroplaneDamageTraceGfmic,self).__init__(**kwrds)
        self._trace_time = []
#        self._trace_Gf   = []

    def bind( self ):
        '''
        Locate the evaluators
        '''
        super(MATS2DMicroplaneDamageTraceGfmic,self).bind()
        # get the evaluator for the variable
        self.var_time_eval = self.rmgr.rte_dict['time']
            
            
    def get_MPW(self):
        '''
        Get the microplanes weighting coefficients used for numerical integration
        '''
        # this is special for the microplane model
        return self.var_x_eval.ts._MPW


    def add_current_values( self, sctx, eps_eng ):
        '''
        Invoke the evaluators in the current context for the specified control vector U_k.
        '''
        super(MATS2DMicroplaneDamageTraceGfmic,self).add_current_values( sctx, eps_eng )
        time_value = self.var_time_eval( sctx, eps_eng )
#        Gf_value   = self.G_f
        self._trace_time.append( time_value[0] )
#        self._trace_Gf.append(Gf_value[0])    
    
    def redraw(self):

        if self.idx_x < 0 or self.idx_y < 0 or \
            self._xdata == [] or self._ydata == []:
            return

        xarray = array( self._xdata )
        yarray = array( self._ydata )

        # get the microplane weighting coefficients:
        MPW  = self.get_MPW()

        # derive the number of specified microplanes base on the shape of the trace 
        n_mp = xarray.shape[1]

        # @todo: find an alternativ (faster) numpy way to calculate the value for G_f 
        #trapz.reduce( ( yarray[0,:], xarray[0,:] )


        # total energy contribution of the microplanes:
        E_t_arr = array([ self._get_E_t( xarray[:,i_mp], yarray[:,i_mp]) for i_mp in range(n_mp)])

        # elastic energy contribution of the microplanes:
        U_t_arr = array([ self._get_U_t( xarray[:,i_mp], yarray[:,i_mp]) for i_mp in range(n_mp)])

        G_f_arr = E_t_arr - U_t_arr
        
        # total "total energy" = weighted sum over the microplanes:
        E_t = array([ dot(E_t_arr, MPW) ])
        print('E_t_mic = %.10f' %(E_t ))

        # total "elastic energy" = weighted sum over the microplanes:
        U_t = array([ dot(U_t_arr, MPW) ])
        print('U_t_mic = %.10f' %( U_t ))

        # total "fracture energy" = weighted sum over the microplanes:
        G_f = array([ dot(G_f_arr, MPW) ])
        print('G_f_mic = %.10f' %( G_f ))
        print('\n')

        self.trace.xdata = array( self._trace_time, dtype = float )
#        self.trace.ydata = array( self._trace_Gf, dtype = float )
        self.trace.ydata = G_f * ones( self.trace.xdata.shape, dtype = float )
                

    def _get_E_t(self, xdata, ydata):
        # integral under the stress strain curve
        E_t = trapz( ydata, xdata )
        return E_t
        
    
    def _get_U_t(self, xdata, ydata):
        # area of the stored elastic energy  
        U_t = 0.0
        if len( xdata ) != 0:
            U_t = 0.5 * ydata[-1] * xdata[-1]        
        return U_t

    
    def clear(self):
        super(MATS2DMicroplaneDamageTraceGfmic,self).clear()
        self._trace_time = []



class MATS2DMicroplaneDamageTraceEtmic( MATS2DMicroplaneDamageTraceGfmic ):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''
    def redraw(self):
        if self.idx_x < 0 or self.idx_y < 0 or \
            self._xdata == [] or self._ydata == []:
            return
        xarray = array( self._xdata )
        yarray = array( self._ydata )
        # get the microplane weighting coefficients:
        MPW  = self.get_MPW()
        # derive the number of specified microplanes base on the shape of the trace 
        n_mp = xarray.shape[1]
        # total energy contribution of the microplanes:
        E_t_arr = array([ self._get_E_t( xarray[:,i_mp], yarray[:,i_mp]) for i_mp in range(n_mp)])
        # total "total energy" = weighted sum over the microplanes:
        E_t = array([ dot(E_t_arr, MPW) ])
        print('E_t_mic = %.10f' %(E_t ))

        self.trace.xdata = array( self._trace_time, dtype = float )
        self.trace.ydata = E_t * ones( self.trace.xdata.shape, dtype = float )
        

    
class MATS2DMicroplaneDamageTraceUtmic( MATS2DMicroplaneDamageTraceGfmic ):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''    
    def redraw(self):
        if self.idx_x < 0 or self.idx_y < 0 or \
            self._xdata == [] or self._ydata == []:
            return

        xarray = array( self._xdata )
        yarray = array( self._ydata )

        # get the microplane weighting coefficients:
        MPW  = self.get_MPW()

        # derive the number of specified microplanes base on the shape of the trace 
        n_mp = xarray.shape[1]

        # elastic energy contribution of the microplanes:
        U_t_arr = array([ self._get_U_t( xarray[:,i_mp], yarray[:,i_mp]) for i_mp in range(n_mp)])
        # total "elastic energy" = weighted sum over the microplanes:
        U_t = array([ dot(U_t_arr, MPW) ])
 
        self.trace.xdata = array( self._trace_time, dtype = float )
        self.trace.ydata = U_t * ones( self.trace.xdata.shape, dtype = float )




from os.path import join
from ibvpy.mathkit.mfn import MFnLineArray
from traits.api import Str, \
    Range, Property, cached_property, observe
from bmcs_utils.api import Float, Int, Enum, FloatRangeEditor
from ibvpy.view.plot2d import Viz2D
from ibvpy.view.ui import BMCSLeafNode
from ibvpy.view.reporter import RInputRecord
import bmcs_utils.api as bu
import numpy as np


class LoadingScenario(MFnLineArray, BMCSLeafNode, RInputRecord):
    """Loading scenario.
    """
    def reset(self):
        return

    node_name = Str('loading scenario')

    t_max = Float(1.)

    xdata = Property
    def _get_xdata(self):
        return self.xy_arrays[0]

    ydata = Property
    def _get_ydata(self):
        return self.xy_arrays[1]

    def update_plot(self, axes):
        axes.plot(self.xdata, self.ydata)
        axes.fill_between(self.xdata, self.ydata, 0, alpha=0.1)


class MonotonicLoadingScenario(LoadingScenario):

    n_incr = Int(10, BC=True)
    maximum_loading = Float(1.0, BC=True,
                            enter_set=True, auto_set=False,
                            symbol='\phi_{\max}',
                            desc='load factor at maximum load level',
                            unit='-')

    ipw_view = bu.View(
        bu.Item("n_incr"),
        bu.Item('maximum_loading'),
    )

    xy_arrays = Property(depends_on="state_changed")
    @cached_property
    def _get_xy_arrays(self):
        t_arr = np.linspace(0, self.t_max, self.n_incr)
        d_arr = np.linspace(0, self.maximum_loading, self.n_incr)
        return t_arr, d_arr

    def write_figure(self, f, rdir, rel_study_path):
        print('FNAME', self.node_name)
        fname = 'fig_' + self.node_name.replace(' ', '_') + '.pdf'
        print('FNAME', fname)
        self._update_xy_arrays()
        f.write(r'''
\multicolumn{3}{r}{\includegraphics[width=5cm]{%s}}\\
''' % join(rel_study_path, fname))
        self.savefig(join(rdir, fname))


class CyclicLoadingScenario(LoadingScenario):

    number_of_cycles = Int(1, BC=True,
                           enter_set=True, auto_set=False,
                           symbol='n_\mathrm{cycles}',
                           unit='-',
                           desc='for cyclic loading',
                           )
    maximum_loading = Float(1.0, BC=True,
                            enter_set=True, auto_set=False,
                            symbol='\phi_{\max}',
                            desc='load factor at maximum load level',
                            unit='-')
    number_of_increments = Int(20, BC=True,
                               enter_set=True, auto_set=False,
                               symbol='n_{\mathrm{incr}}',
                               unit='-',
                               desc='number of values within a monotonic load branch')
    unloading_ratio = Float(0.5, BC=True,
                            enter_set=True, auto_set=False,
                            symbol='\phi_{\mathrm{unload}}',
                            desc='fraction of maximum load at lowest load level',
                            unit='-')
    amplitude_type = Enum(options=["increasing", "constant"],
                          enter_set=True, auto_set=False,
                          symbol='option',
                          unit='-',
                          desc='possible values: [increasing, constant]',
                          BC=True)
    loading_range = Enum(options=["non-symmetric", "symmetric"],
                         enter_set=True, auto_set=False,
                         symbol='option',
                         unit='-',
                         desc='possible values: [non-symmetric, symmetric]',
                         BC=True)

    ipw_view = bu.View(
        bu.Item('number_of_cycles'),
        bu.Item('maximum_loading'),
        bu.Item('number_of_increments'),
        bu.Item('unloading_ratio'), # , editor=FloatRangeEditor(low=0, high=1)),
        bu.Item('amplitude_type'),
        bu.Item('loading_range'),
    )
    xy_arrays = Property(depends_on="state_changed")
    @cached_property
    def _get_xy_arrays(self):
        if(self.amplitude_type == "increasing" and
                self.loading_range == "symmetric"):
            d_levels = np.linspace(
                0, self.maximum_loading, self.number_of_cycles * 2)
            d_levels.reshape(-1, 2)[:, 0] *= -1
            d_history = d_levels.flatten()
            d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1],
                                           self.number_of_increments)
                               for i in range(len(d_levels) - 1)])

        if(self.amplitude_type == "increasing" and
                self.loading_range == "non-symmetric"):
            d_levels = np.linspace(
                0, self.maximum_loading, self.number_of_cycles * 2)
            d_levels.reshape(-1, 2)[:, 0] *= 0
            d_history = d_levels.flatten()
            d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1],
                                           self.number_of_increments)
                               for i in range(len(d_levels) - 1)])

        if(self.amplitude_type == "constant" and
                self.loading_range == "symmetric"):
            d_levels = np.linspace(
                0, self.maximum_loading, self.number_of_cycles * 2)
            d_levels.reshape(-1, 2)[:, 0] = -self.maximum_loading
            d_levels[0] = 0
            d_levels.reshape(-1, 2)[:, 1] = self.maximum_loading
            d_history = d_levels.flatten()
            d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                               for i in range(len(d_levels) - 1)])

        if(self.amplitude_type == "constant" and
                self.loading_range == "non-symmetric"):
            d_levels = np.linspace(
                0, self.maximum_loading, self.number_of_cycles * 2)
            d_levels.reshape(-1, 2)[:,
                                    0] = self.maximum_loading * self.unloading_ratio
            d_levels[0] = 0
            d_levels.reshape(-1, 2)[:, 1] = self.maximum_loading
            d_history = d_levels.flatten()
            d_arr = np.hstack([np.linspace(d_history[i], d_history[i + 1], self.number_of_increments)
                               for i in range(len(d_levels) - 1)])

        t_arr = np.linspace(0, self.t_max, len(d_arr))
        return t_arr, d_arr

class Viz2DLoadControlFunction(Viz2D):
    '''Plot adaptor for the pull-out simulator.
    '''
    label = 'Load control'

    def plot(self, ax, vot, *args, **kw):
        bc = self.vis2d.control_bc
        val = bc.value
        tloop = self.vis2d.tloop
        t_arr = np.array(tloop.t_record, np.float64)
        if len(t_arr) == 0:
            return
        f_arr = val * bc.time_function(t_arr)
        ax.plot(t_arr, f_arr, color='black')
        ax.set_ylabel('load factor')
        ax.set_xlabel('time')
        self.plot_marker(ax, vot)

    def plot_marker(self, ax, vot):
        bc = self.vis2d.control_bc
        val = bc.value
        tloop = self.vis2d.tloop
        t_arr = np.array(tloop.t_record, np.float64)
        if len(t_arr) == 0:
            return
        f_arr = val * bc.time_function(t_arr)
        vot_idx = tloop.get_time_idx(vot)
        ax.plot([t_arr[vot_idx]], [f_arr[vot_idx]], 'o',
                color='black', markersize=10)

    def plot_tex(self, ax, vot):
        self.plot(ax, vot)




from .matsXD_cmdm_phi_fn import \
    PhiFnStrainSoftening, PhiFnGeneral, PhiFnGeneralExtended, \
    PhiFnGeneralExtendedExp,  PhiFnStrainHardening, PhiFnStrainHardeningBezier, \
    PhiFnStrainHardeningLinear


#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on May 26, 2009 by: rchx

from traits.api import  \
    Str, Enum, \
    List, cached_property, \
    provides, Property
from .i_bcond import \
    IBCond
from ibvpy.view.plot2d import Vis2D
from ibvpy.view.ui import BMCSTreeNode

from .bc_dof import BCDof


@provides(IBCond)
class BCDofList(BMCSTreeNode, Vis2D):
    '''
    Implements the IBC functionality for a constrained dof.
    '''
    tree_node_list = List

    tree_node_list = Property(depends_on='bcdof_list,bcdof_list_items')

    @cached_property
    def _get_tree_node_list(self):
        return self.bcdof_list

    name = Str('<unnamed>')

    node_name = Property

    def _get_node_name(self):
        s = '%s:%s=%s' % (self.var, self.slice, self.value)
        return s

    var = Enum('u', 'f', 'eps', 'sig')

    def is_essential(self):
        return self.var == 'u'

    def is_linked(self):

        return self.link_dofs != []

    def is_constrained(self):
        '''
        Return true if a DOF is either explicitly prescribed or it depends on other DOFS.
        '''
        return self.is_essential() or self.is_linked()

    def is_natural(self):
        return self.var == 'f' or self.var == 'eps' or self.var == 'sig'

    bcdof_list = List(BCDof)

    def reset(self):
        self.bcdof_list = []

    integ_domain = Enum(['global', 'local'])

    def setup(self, sctx):
        '''
        Locate the spatial context.f
        '''

    def apply_essential(self, K):

        for bcond in self.bcdof_list:
            bcond.apply_essential(K)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):

        if self.is_essential():
            for bcond in self.bcdof_list:
                bcond.apply(step_flag, sctx, K, R, t_n, t_n1)
        else:
            self.apply_natural(step_flag, sctx, K, R, t_n, t_n1)

    def apply_natural(self, step_flag, sctx, K, R, t_n, t_n1):

        raise NotImplementedError




from ibvpy.fets.i_fets import IFETSEval
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray
from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    Property, cached_property, \
    provides, Callable, \
    Tuple, Int, Type, Array, Float, Instance, Bool, DelegatesTo
from ibvpy.view.ui.bmcs_tree_node import BMCSTreeNode
from ibvpy.mesh.i_fe_uniform_domain import IFEUniformDomain

import numpy as np


@provides(IXModel)
class XDomainFE(BMCSTreeNode):

    hidden = Bool(False)
    #=========================================================================
    # Type and shape specification of state variables representing the domain
    #=========================================================================
    U_var_shape = Property(Int)

    def _get_U_var_shape(self):
        return self.mesh.n_dofs

    vtk_expand_operator = Property

    def _get_vtk_expand_operator(self):
        return self.fets.vtk_expand_operator

    K_type = Type(SysMtxArray)

    state_var_shape = Property(Tuple)

    def _get_state_var_shape(self):
        return (self.mesh.n_active_elems, self.fets.n_m,)

    #=========================================================================
    # Methods needed by XDomain to chain the subdomains
    #=========================================================================
    dof_offset = Property

    def _get_dof_offset(self):
        return self.mesh.dof_offset

    n_active_elems = Property

    def _get_n_active_elems(self):
        return self.mesh.n_active_elems

    def set_next(self, next_):
        self.mesh.next_grid = next_.mesh

    def set_prev(self, prev):
        self.mesh.prev_grid = prev.mesh

    #=========================================================================
    # Input parameters
    #=========================================================================
    coord_min = Array(Float, value=[0., 0., 0.], GEO=True)
    '''Grid geometry specification - min corner point
    '''
    coord_max = Array(Float, value=[1., 1., 1.], MESH=True)
    '''Grid geometry specification - max corner point
    '''
    shape = Array(Int, value=[1, 1, 1], MESH=True)
    '''Number of elements in the individual dimensions
    '''
    geo_transform = Callable
    '''Geometry transformation
    '''
    integ_factor = Float(1.0, input=True, CS=True)
    '''Integration factor used to multiply the integral
    '''
    fets = Instance(IFETSEval, input=True, FE=True)
    '''Finite element type
    '''

    dim_u = Int(2)

    Diff1_abcd = Array(np.float64, input=True)
    '''Symmetric operator distributing the first order
    derivatives of the shape functions into the 
    tensor field
    '''

    def _Diff1_abcd_default(self):
        delta = np.identity(self.dim_u)
        # symmetrization operator
        Diff1_abcd = 0.5 * (
            np.einsum('ac,bd->abcd', delta, delta) +
            np.einsum('ad,bc->abcd', delta, delta)
        )
        return Diff1_abcd

    mesh = Instance(IFEUniformDomain)

    X_Id = DelegatesTo('mesh')
    I_Ei = DelegatesTo('mesh')

    x_Eia = Property(depends_on='MESH,GEO,CS,FE')

    def _get_x_Eia(self):
        x_Ia = self.X_Id
        I_Ei = self.I_Ei
        x_Eia = x_Ia[I_Ei, :]
        return x_Eia

    x_Ema = Property(depends_on='MESH,GEO,CS,FE')

    def _get_x_Ema(self):
        return np.einsum(
            'im,Eia->Ema', self.fets.N_im, self.x_Eia
        )

    o_Ia = Property(depends_on='MESH,GEO,CS,FE')

    @cached_property
    def _get_o_Ia(self):
        x_Ia = self.mesh.X_Id
        n_I, _ = x_Ia.shape
        n_a = self.mesh.n_nodal_dofs
        do = self.mesh.dof_offset
        return do + np.arange(n_I * n_a, dtype=np.int_).reshape(-1, n_a)

    o_Eia = Property(depends_on='MESH,GEO,CS,FE')

    @cached_property
    def _get_o_Eia(self):
        I_Ei = self.I_Ei
        return self.o_Ia[I_Ei]

    B1_Einabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacement and strain in every
    visualization point
    '''

    @cached_property
    def _get_B1_Einabc(self):
        inv_J_Enar = np.linalg.inv(self.J_Enar)
        print('************************************* inv_J', inv_J_Enar)
        return np.einsum(
            'abcd,imr,Eidr->Eimabc',
#            'abcd,imr,Emrd->Eimabc',
            self.Diff1_abcd, self.fets.dN_inr, inv_J_Enar
        )

    I_Ei = Property(depends_on='MESH,GEO,CS,FE')
    '''[element, node] -> global node
    '''

    def _get_I_Ei(self):
        return self.mesh.I_Ei

    det_J_Em = Property(depends_on='MESH,GEO,CS,FE')
    '''Jacobi matrix in integration points
    '''
    @cached_property
    def _get_det_J_Em(self):
        return np.linalg.det(self.J_Emar)

    J_Emar = Property(depends_on='MESH,GEO,CS,FE')
    '''Jacobi matrix in integration points
    '''
    @cached_property
    def _get_J_Emar(self):
        return np.einsum(
            'imr,Eia->Emar', self.fets.dN_imr, self.x_Eia
        )

    J_Enar = Property(depends_on='MESH,GEO,CS,FE')
    '''Jacobi matrix in nodal points
    '''
    @cached_property
    def _get_J_Enar(self):
        return np.einsum(
            'inr,Eia->Enar',
            self.fets.dN_inr, self.x_Eia
        )

    #=========================================================================
    # Conversion between linear algebra objects and field variables
    #=========================================================================
    B1_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B1_Eimabc(self):
        inv_J_Emar = np.linalg.inv(self.J_Emar)
        return np.einsum(
            'abcd,inr,Emrd->Einabc',
            self.Diff1_abcd, self.fets.dN_imr, inv_J_Emar
        )

    B_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B_Eimabc(self):
        return self.B1_Eimabc

    BB_Emicjdabef = Property(depends_on='MESH,GEO,CS,FE')
    '''Quadratic form of the kinematic mapping.
    '''

    def _get_BB_Emicjdabef(self):
        return np.einsum(
            '...Eimabc,...Ejmefd, Em, m->...Emicjdabef',
            self.B_Eimabc, self.B_Eimabc, self.det_J_Em, self.fets.w_m
        )

    n_dofs = Property

    def _get_n_dofs(self):
        return self.mesh.n_dofs

    def U2u(self, U_Eia):
        return U_Eia

    def f2F(self, f_Eic):
        return f_Eic

    def k2K(self, K_Eicjd):
        return K_Eicjd

    def map_U_to_field(self, U):
        U_Eia = U[self.o_Eia]
        # coordinate transform to local
        U_Eia = self.xU2u(U_Eia)
        eps_Emab = np.einsum(
            'Eimabc,Eic->Emab',
            self.B_Eimabc, U_Eia
        )
        return eps_Emab

    def map_field_to_F(self, sig_Emab):
        f_Eic = self.integ_factor * np.einsum(
            'm,Eimabc,Emab,Em->Eic',
            self.fets.w_m, self.B_Eimabc, sig_Emab, self.det_J_Em
        )
        # coordinate transform to global
        f_Eic = self.xf2F(f_Eic)
        _, n_i, n_c = f_Eic.shape
        f_Ei = f_Eic.reshape(-1, n_i * n_c)
        o_E = self.o_Eia.reshape(-1, n_i * n_c)
        return o_E.flatten(), f_Ei.flatten()

    def map_field_to_K(self, D_Emabef):
        k_Eicjd = self.integ_factor * np.einsum(
            'Emicjdabef,Emabef->Eicjd',
            self.BB_Emicjdabef, D_Emabef
        )
        _, _, n_a, n_b, _,_ = D_Emabef.shape
        # coordinate transform to global
        _K_Eicjd = self.k2K(k_Eicjd)
        K_Eicjd = self.xk2K(k_Eicjd)
        # print(_K_Eicjd - K_Eicjd)
        _, n_i, n_c, n_j, n_d = K_Eicjd.shape
        K_Eij = K_Eicjd.reshape(-1, n_i * n_c, n_j * n_d)
        o_Ei = self.o_Eia.reshape(-1, n_i * n_c)
        #print(K_Eij)
        return SysMtxArray(mtx_arr=K_Eij, dof_map_arr=o_Ei)

    debug_cell_data = Bool(False)
    # @todo - comment this procedure`

    def get_vtk_cell_data(self, position, point_offset, cell_offset):
        if position == 'nodes':
            subcell_offsets, subcell_lengths, subcells, subcell_types = \
                self.fets.vtk_node_cell_data
        elif position == 'int_pnts':
            subcell_offsets, subcell_lengths, subcells, subcell_types = \
                self.fets.vtk_ip_cell_data

        if self.debug_cell_data:
            print('subcell_offsets')
            print(subcell_offsets)
            print('subcell_lengths')
            print(subcell_lengths)
            print('subcells')
            print(subcells)
            print('subcell_types')
            print(subcell_types)

        n_subcells = subcell_types.shape[0]
        n_cell_points = self.n_cell_points
        subcell_size = subcells.shape[0] + n_subcells

        if self.debug_cell_data:
            print('n_cell_points', n_cell_points)
            print('n_cells', self.n_cells)

        vtk_cell_array = np.zeros((self.n_cells, subcell_size), dtype=int)

        idx_cell_pnts = np.repeat(True, subcell_size)

        if self.debug_cell_data:
            print('idx_cell_pnts')
            print(idx_cell_pnts)

        idx_cell_pnts[subcell_offsets] = False

        if self.debug_cell_data:
            print('idx_cell_pnts')
            print(idx_cell_pnts)

        idx_lengths = idx_cell_pnts == False

        if self.debug_cell_data:
            print('idx_lengths')
            print(idx_lengths)

        point_offsets = np.arange(self.n_cells) * n_cell_points
        point_offsets += point_offset

        if self.debug_cell_data:
            print('point_offsets')
            print(point_offsets)

        vtk_cell_array[:, idx_cell_pnts] = point_offsets[
            :, None] + subcells[None, :]
        vtk_cell_array[:, idx_lengths] = subcell_lengths[None, :]

        if self.debug_cell_data:
            print('vtk_cell_array')
            print(vtk_cell_array)

        n_active_cells = self.mesh.n_active_elems

        if self.debug_cell_data:
            print('n active cells')
            print(n_active_cells)

        cell_offsets = np.arange(n_active_cells, dtype=int) * subcell_size
        cell_offsets += cell_offset
        vtk_cell_offsets = cell_offsets[:, None] + subcell_offsets[None, :]

        if self.debug_cell_data:
            print('vtk_cell_offsets')
            print(vtk_cell_offsets)

        vtk_cell_types = np.zeros(
            self.n_cells * n_subcells, dtype=int
        ).reshape(self.n_cells, n_subcells)
        vtk_cell_types += subcell_types[None, :]

        if self.debug_cell_data:
            print('vtk_cell_types')
            print(vtk_cell_types)

        return (vtk_cell_array.flatten(),
                vtk_cell_offsets.flatten(),
                vtk_cell_types.flatten())

    n_cells = Property(Int)

    def _get_n_cells(self):
        '''Return the total number of cells'''
        return self.mesh.n_active_elems

    n_cell_points = Property(Int)

    def _get_n_cell_points(self):
        '''Return the number of points defining one cell'''
        return self.fets.n_vtk_r

@provides(IXModel)
class XDomainFEGrid(XDomainFE):

    #=========================================================================
    # Finite element discretization respecting the FE definition
    #=========================================================================
    mesh = Property(Instance(FEGrid), depends_on='MESH,GEO')

    @cached_property
    def _get_mesh(self):
        return FEGrid(coord_min=self.coord_min,
                      coord_max=self.coord_max,
                      shape=self.shape,
                      geo_transform=self.geo_transform,
                      fets_eval=self.fets)


if __name__ == '__main__':
    from ibvpy.fets.fets1D5.fets1d52ulrhfatigue import FETS1D52ULRHFatigue
    xd = XDomainFEGrid(coord_max=(1,), shape=(1,),
                       dim_u=2, fets=FETS1D52ULRHFatigue())
    print(xd.BB_Emicjdabef.shape)
    print(xd.get_vtk_cell_data('nodes', 0, 0))



'''
Created on Mar 4, 2017

@author: rch
'''

from traits.api import \
    List, Float, Int, Range, Property,\
    cached_property

import numpy as np
from ibvpy.view.plot2d import Vis2D, Viz2D
from ibvpy.view.ui.bmcs_tree_node import BMCSLeafNode


class TFViz2D(Viz2D):

    def plot(self, ax, vot):
        t, y = self.vis2d.get_ty_data(vot)
        ax.plot(t, y)
        y_min, y_max = np.min(y), np.max(y)
        ax.plot([vot, vot], [y_min, y_max])


class BoundaryCondition(BMCSLeafNode, Vis2D):

    node_name = 'boundary condition'
    t_values = List(Float, [0])
    f_values = List(Float, [0])
    n_f_values = Int(10, auto_set=False, enter_set=True)
    f_min = Float(0.0, auto_set=False, enter_set=True)
    f_max = Float(1.0, auto_set=False, enter_set=True)
    t_ref = Float(1.0, auto_set=False, enter_set=True)

    f_value = Range(low='f_min', high='f_max', value=0,
                    auto_set=False, enter_set=True)

    d_t = Property(depends_on='t_ref,n_f_values')

    @cached_property
    def _get_d_t(self):
        return self.t_ref / self.n_f_values

    def _f_value_changed(self):
        delta_f = self.f_value - self.f_values[-1]
        self.f_values.append(self.f_value)
        rel_step = delta_f / self.f_max
        delta_t = rel_step * self.t_ref
        t_value = np.fabs(delta_t) + self.t_values[-1]
        self.t_values.append(t_value)
        if self.ui:
            self.ui.set_vot(t_value)

    def get_ty_data(self, vot):
        return self.t_values, self.f_values

    viz2d_classes = {
        'time_function': TFViz2D,
    }




'''

'''

from os.path import join

from ibvpy.tmodel import MATSEval
from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, GfDamageFn, ExpSlopeDamageFn, AbaqusDamageFn, \
    LinearDamageFn, FRPDamageFn, WeibullDamageFn
from ibvpy.mathkit.mfn.mfn_line.mfn_line import MFnLineArray
from traits.api import  \
    observe, List, on_trait_change, \
    Instance, Trait, Bool, Property, cached_property
import bmcs_utils.api as bu
import numpy as np

class MATSEval1D5(MATSEval):
    """Base class defining the plot
    """
    s_max = bu.Float(0.5, tooltip='Visualization limit [m]')

    def subplots(self, fig):
        ax_tau = fig.subplots(1,1)
        ax_d_tau = ax_tau.twinx()
        return ax_tau, ax_d_tau

    def update_plot(self, axes):
        ax_tau, ax_d_tau = axes
        s_max = self.s_max
        n_s = 100
        s_range = np.linspace(1e-9,s_max,n_s)
        eps_range = np.zeros((n_s, 3))
        eps_range[:,1] = s_range
        state_vars = { var : np.zeros( (n_s,) + shape )
            for var, shape in self.state_var_shapes.items()
        }
        sig_range, D = self.get_corr_pred(eps_range, 1, **state_vars)
        tau_range = sig_range[:,1]
        ax_tau.plot(s_range, tau_range,color='blue')
        d_tau_range = D[...,1,1]
        ax_d_tau.plot(s_range, d_tau_range,
                      linestyle='dashed', color='gray')
        ax_tau.set_xlabel(r'$s$ [mm]')
        ax_tau.set_ylabel(r'$\tau$ [MPa]')
        ax_d_tau.set_ylabel(r'$\mathrm{d} \tau / \mathrm{d} s$ [MPa/mm]')

class MATS1D5BondSlipMultiLinear(MATSEval1D5):
    """Multilinear bond-slip law
    """
    name = "multilinear bond law"

    E_m = bu.Float(28000.0, tooltip='Stiffness of the matrix [MPa]',
                MAT=True, unit='MPa', symbol='E_\mathrm{m}',
                desc='E-modulus of the matrix')

    E_f = bu.Float(170000.0, tooltip='Stiffness of the fiber [MPa]',
                MAT=True, unit='MPa', symbol='E_\mathrm{f}',
                desc='E-modulus of the reinforcement')

    s_data = bu.Str('0,1', tooltip='Comma-separated list of strain values',
                 MAT=True, unit='mm', symbol='s',
                 desc='slip values')

    tau_data = bu.Str('0,1', tooltip='Comma-separated list of stress values',
                   MAT=True, unit='MPa', symbol=r'\tau',
                   desc='shear stress values')

    s_tau_table = Property(depends_on='state_changed')
    @cached_property
    def _get_s_tau_table(self):
        s_data = np.fromstring(self.s_data, dtype=np.float64, sep=',')
        tau_data = np.fromstring(self.tau_data, dtype=np.float64, sep=',')
        if len(s_data) != len(tau_data):
            raise ValueError('s array and tau array must have the same size')
        return s_data, tau_data

    #=========================================================================
    # Configurational parameters
    #=========================================================================
    U_var_shape = (1,)
    '''Shape of the primary variable required by the TStepState.
    '''

    state_var_shapes = {}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    node_name = 'multiply linear bond'

    def get_corr_pred(self, eps_n1, t_n1):
        D_shape = eps_n1.shape[:-1] + (3, 3)
        D = np.zeros(D_shape, dtype=np.float64)

        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f

        tau = np.einsum('...st,...t->...s', D, eps_n1)
        s = eps_n1[..., 1]
        shape = s.shape
        signs = np.sign(s.flatten())
        s_pos = np.fabs(s.flatten())
        bs_law = self.bs_law
        tau[..., 1] = (signs * bs_law(s_pos)).reshape(*shape)
        D_tau = self.bs_law.diff(s_pos).reshape(*shape)
        D[..., 1, 1] = D_tau

        return tau, D

    bs_law = Property(depends_on='state_changed')
    @cached_property
    def _get_bs_law(self):
        s_data, tau_data = self.s_tau_table
        return MFnLineArray(
            xdata=s_data,
            ydata=tau_data,
            plot_diff=False
        )

    ipw_view = bu.View(
        bu.Item('E_m'),
        bu.Item('E_f'),
        bu.Item('s_data'),
        bu.Item('tau_data'),
    )

class MATS1D5BondSlipD(MATSEval1D5):
    name = 'damage model'
    node_name = 'damage model'

    E_m = bu.Float(30000.0, tooltip='Stiffness of the matrix [MPa]',
                MAT=True)

    E_f = bu.Float(200000.0, tooltip='Stiffness of the fiber [MPa]',
                MAT=True)

    E_b = bu.Float(10000.0, tooltip='Stiffness of the fiber [MPa]',
                MAT=True)

    omega_fn = bu.EitherType(
        options=[('linear', LinearDamageFn),
                 ('abaqus', AbaqusDamageFn),
                 ('exp-slope', ExpSlopeDamageFn),
                 ('fracture-energy', GfDamageFn),
                 ('weibull-CDF', WeibullDamageFn),
                 ],
        MAT=True,
        on_option_change='link_omega_to_mats'
    )

    D_alg = bu.Float(0, MAT=True)

    # upon change of the type attribute set the link to the material model
    def link_omega_to_mats(self):
        self.omega_fn_.trait_set(mats=self,
                                 E_name='E_b',
                                 x_max_name='s_max')

    tree = ['omega_fn']

    def omega(self, k):
        return self.omega_fn_(k)

    def omega_derivative(self, k):
        return self.omega_fn_.diff(k)

    state_var_shapes = {
        'kappa_n' : (),
        'omega_n' : ()
    }

    def get_corr_pred(self, eps_n1, t_n1, kappa_n, omega_n):
        D_shape = eps_n1.shape[:-1] + (3, 3)
        D = np.zeros(D_shape, dtype=np.float64)
        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f
        s_n1 = eps_n1[..., 1]
        kappa_n[...] = np.max(np.array([kappa_n, np.fabs(s_n1)]), axis=0)
        omega_n[...] = self.omega(kappa_n)
        tau = np.einsum('...st,...t->...s', D, eps_n1)
        tau[..., 1] = (1 - omega_n) * self.E_b * s_n1
        D[..., 1, 1] = (1 - omega_n) * self.E_b
        if self.D_alg > 0:
            I = self.omega_fn_.get_f_trial(np.fabs(s_n1), kappa_n)
            domega_ds_I = self.omega_derivative(kappa_n[I])
            D_red_I = domega_ds_I * np.fabs(s_n1[I]) * self.E_b * self.D_alg
            D[I+(1, 1)] -= D_red_I
        return tau, D

    ipw_view = bu.View(
        bu.Item('E_m', latex=r'E_\mathrm{m}'),
        bu.Item('E_f', latex=r'E_\mathrm{f}'),
        bu.Item('E_b', latex=r'E_\mathrm{b}'),
        bu.Item('omega_fn', latex=r'\omega(s)'),
        bu.Item('D_alg', latex=r'\theta_\mathrm{alg. stiff.}',
                editor=bu.FloatRangeEditor(low=0,high=1)),
        bu.Item('s_max')
    )

class MATSBondSlipDP(MATSEval1D5):

    node_name = 'bond model: damage-plasticity'

    tree_node_list = List([])

    def _tree_node_list_default(self):
        return [self.omega_fn, ]

    @on_trait_change('omega_fn_type')
    def _update_node_list(self):
        self.tree_node_list = [self.omega_fn]

    E_m = bu.Float(30000.0, tooltip='Stiffness of the matrix [MPa]',
                symbol='E_\mathrm{m}',
                unit='MPa',
                desc='Stiffness of the matrix',
                MAT=True,
                auto_set=True, enter_set=True)

    E_f = bu.Float(200000.0, tooltip='Stiffness of the reinforcement [MPa]',
                symbol='E_\mathrm{f}',
                unit='MPa',
                desc='Stiffness of the reinforcement',
                MAT=True,
                auto_set=False, enter_set=False)

    E_b = bu.Float(12900.0,
                symbol="E_\mathrm{b}",
                unit='MPa',
                desc="Bond stiffness",
                MAT=True,
                enter_set=True,
                auto_set=False)

    gamma = bu.Float(100.0,
                  symbol="\gamma",
                  unit='MPa',
                  desc="Kinematic hardening modulus",
                  MAT=True,
                  enter_set=True,
                  auto_set=False)

    K = bu.Float(1000.0,
              symbol="K",
              unit='MPa',
              desc="Isotropic hardening modulus",
              MAT=True,
              enter_set=True,
              auto_set=False)

    tau_bar = bu.Float(5.0,
                    symbol=r'\bar{\tau}',
                    unite='MPa',
                    desc="Reversibility limit",
                    MAT=True,
                    enter_set=True,
                    auto_set=False)

    uncoupled_dp = Bool(False,
                        MAT=True,
                        label='Uncoupled d-p'
                        )

    s_0 = bu.Float(MAT=True,
                desc='Elastic strain/displacement limit')

    def __init__(self, *args, **kw):
        super(MATSBondSlipDP, self).__init__(*args, **kw)
        self._omega_fn_type_changed()
        self._update_s0()

    @on_trait_change('tau_bar,E_b')
    def _update_s0(self):
        if not self.uncoupled_dp:
            if self.E_b == 0:
                self.s_0 = 0
            else:
                self.s_0 = self.tau_bar / self.E_b
            self.omega_fn.s_0 = self.s_0

    omega_fn_type = Trait('multilinear',
                          dict(selfregularized=GfDamageFn,
                               exp_slope=ExpSlopeDamageFn,
                               abaqus=AbaqusDamageFn,
                               # FRP=FRPDamageFn,
                               # multilinear=MultilinearDamageFn
                               ),
                          MAT=True,
                          )

    def _omega_fn_type_changed(self):
        self.omega_fn = self.omega_fn_type_(mats=self, s_0=self.s_0)

    omega_fn = Instance(IDamageFn,
                        report=True)

    def _omega_fn_default(self):
        return MultilinearDamageFn()

    def omega(self, k):
        return self.omega_fn(k)

    def omega_derivative(self, k):
        return self.omega_fn.diff(k)

    state_var_shapes = dict(s_pl_n=(),
                            alpha_n=(),
                            z_n=(),
                            kappa_n=(),
                            omega_n=())

    def get_corr_pred(self, eps_n1, t_n1,
                      s_pl_n, alpha_n, z_n, kappa_n, omega_n):

        D_shape = eps_n1.shape[:-1] + (3, 3)
        D = np.zeros(D_shape, dtype=np.float64)
        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f

        s_n1 = eps_n1[..., 1]

        sig_pi_trial = self.E_b * (s_n1 - s_pl_n)

        Z = self.K * z_n

        # for handeling the negative values of isotropic hardening
        h_1 = self.tau_bar + Z
        pos_iso = h_1 > 1e-6

        X = self.gamma * alpha_n

        # for handeling the negative values of kinematic hardening (not yet)
        # h_2 = h * np.sign(sig_pi_trial - X) * \
        #    np.sign(sig_pi_trial) + X * np.sign(sig_pi_trial)
        #pos_kin = h_2 > 1e-6

        f_trial = np.fabs(sig_pi_trial - X) - h_1 * pos_iso

        I = f_trial > 1e-6

        tau = np.einsum('...st,...t->...s', D, eps_n1)
        # Return mapping
        delta_lamda_I = f_trial[I] / (self.E_b + self.gamma + np.fabs(self.K))

        # update all the state variables
        s_pl_n[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])
        z_n[I] += delta_lamda_I
        alpha_n[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])

        kappa_n[I] = np.max(
            np.array([kappa_n[I], np.fabs(s_n1[I])]), axis=0)
        omega_n[I] = self.omega(kappa_n[I])

        tau[..., 1] = (1 - omega_n) * self.E_b * (s_n1 - s_pl_n)

        domega_ds_I = self.omega_derivative(kappa_n[I])

        # Consistent tangent operator
        D_ed_I = -self.E_b / (self.E_b + self.K + self.gamma) \
            * domega_ds_I * self.E_b * (s_n1[I] - s_pl_n[I]) \
            + (1 - omega_n[I]) * self.E_b * (self.K + self.gamma) / \
            (self.E_b + self.K + self.gamma)

        D[..., 1, 1] = (1 - omega_n) * self.E_b
        D[I, 1, 1] = D_ed_I

        return tau, D

    tree_view = bu.View(
        bu.Item('E_m'),
        bu.Item('E_f'),
        bu.Item('E_b'),
        bu.Item('gamma'),
        bu.Item('K'),
        bu.Item('tau_bar'),
        bu.Item('uncoupled_dp'),
        bu.Item('s_0'),  # , enabled_when='uncoupled_dp'),
    )


class MATS1D5BondSlipEP(MATSEval1D5):
    '''Elastc plastic bond slip model
    '''
    name = 'elastic-plastic model'

    E_m = bu.Float(30000.0, tooltip='Stiffness of the matrix [MPa]',
                symbol='E_\mathrm{m}',
                unit='MPa',
                desc='Stiffness of the matrix',
                MAT=True,
                auto_set=True, enter_set=True)

    E_f = bu.Float(200000.0, tooltip='Stiffness of the reinforcement [MPa]',
                symbol='E_\mathrm{f}',
                unit='MPa',
                desc='Stiffness of the reinforcement',
                MAT=True,
                auto_set=False, enter_set=False)

    E_b = bu.Float(12900.0,
                symbol="E_\mathrm{b}",
                unit='MPa',
                desc="Bond stiffness",
                MAT=True,
                enter_set=True,
                auto_set=False)

    gamma = bu.Float(100.0,
                  symbol="\gamma",
                  unit='MPa',
                  desc="Kinematic hardening modulus",
                  MAT=True,
                  enter_set=True,
                  auto_set=False)

    K = bu.Float(1000.0,
              symbol="K",
              unit='MPa',
              desc="Isotropic hardening modulus",
              MAT=True,
              enter_set=True,
              auto_set=False)

    tau_bar = bu.Float(5.0,
                    symbol=r'\bar{\tau}',
                    unite='MPa',
                    desc="Reversibility limit",
                    MAT=True,
                    enter_set=True,
                    auto_set=False)

    state_var_shapes = dict(s_pl_n=(),
                            alpha_n=(),
                            z_n=())

    def get_corr_pred(self, eps_n1, t_n1, s_pl_n, alpha_n, z_n):

        D_shape = eps_n1.shape[:-1] + (3, 3)
        D = np.zeros(D_shape, dtype=np.float64)
        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f

        s_pl_n0 = np.copy(s_pl_n)
        z_n0 = np.copy(z_n)
        alpha_n0 = np.copy(alpha_n)

        s_n1 = eps_n1[..., 1]
        sig_pi_trial = self.E_b * (s_n1 - s_pl_n)
        Z = self.K * z_n
        H_1 = self.tau_bar + Z
        X = self.gamma * alpha_n

        f_trial = np.fabs(sig_pi_trial - X) - H_1

        I = f_trial > 1e-6

        tau = np.einsum('...st,...t->...s', D, eps_n1)
        # Return mapping
        delta_lamda_I = f_trial[I] / (self.E_b + self.gamma + self.K)

        # update all the state variables
        s_pl_n[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])
        z_n[I] += delta_lamda_I
        alpha_n[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])

        tau[..., 1] = self.E_b * (s_n1 - s_pl_n)

        # Consistent tangent operator
        D_ed_I = (self.E_b * (self.K + self.gamma) /
                  (self.E_b + self.K + self.gamma)
                  )

        D[..., 1, 1] = self.E_b
        D[I, 1, 1] = D_ed_I

        Z = self.K * z_n
        # check if the size of the elastic domain is still nonzero
        J = self.tau_bar + Z < 0
        tau[J, 1] = 0
        D[J, 1, 1] = 0
        return tau, D

    ipw_view = bu.View(
        bu.Item('E_m'),
        bu.Item('E_f'),
        bu.Item('E_b'),
        bu.Item('gamma'),
        bu.Item('K'),
        bu.Item('tau_bar'),
    )

    @observe('state_changed')
    def _reset_s_max(self, event=None):
        self.s_max = self.tau_bar / self.E_b * 10

class MATSBondSlipFatigue(MATSEval1D5):

    node_name = 'bond model: bond fatigue'

    E_m = bu.Float(30000, tooltip='Stiffness of the matrix [MPa]',
                MAT=True,
                auto_set=True, enter_set=True)

    E_f = bu.Float(200000, tooltip='Stiffness of the fiber [MPa]',
                MAT=True,
                auto_set=False, enter_set=False)

    E_b = bu.Float(12900,
                label="E_b",
                desc="Bond Stiffness",
                MAT=True,
                enter_set=True,
                auto_set=False)

    gamma = bu.Float(55.0,
                  label="Gamma",
                  desc="Kinematic hardening modulus",
                  MAT=True,
                  enter_set=True,
                  auto_set=False)

    K = bu.Float(11.0,
              label="K",
              desc="Isotropic harening",
              MAT=True,
              enter_set=True,
              auto_set=False)

    S = bu.Float(0.00048,
              label="S",
              desc="Damage cumulation parameter",
              enter_set=True,
              MAT=True,
              auto_set=False)

    r = bu.Float(0.5,
              label="r",
              desc="Damage cumulation parameter",
              MAT=True,
              enter_set=True,
              auto_set=False)

    c = bu.Float(2.8,
              label="c",
              desc="Damage cumulation parameter",
              MAT=True,
              enter_set=True,
              auto_set=False)

    tau_pi_bar = bu.Float(4.2,
                       label="Tau_pi_bar",
                       desc="Reversibility limit",
                       MAT=True,
                       enter_set=True,
                       auto_set=False)

    pressure = bu.Float(0,
                     label="Pressure",
                     desc="Lateral pressure",
                     MAT=True,
                     enter_set=True,
                     auto_set=False)

    a = bu.Float(1.7,
              label="a",
              desc="Lateral pressure coefficient",
              MAT=True,
              enter_set=True,
              auto_set=False)

    state_var_shapes = dict(xs_pi=(),
                            alpha=(),
                            z=(),
                            kappa=(),
                            omega=()
                            )

    def get_corr_pred(self, eps_n1, t_n1, xs_pi, alpha, z, kappa, omega):

        D_shape = eps_n1.shape[:-1] + (3, 3)
        D = np.zeros(D_shape, dtype=np.float64)
        D[..., 0, 0] = self.E_m
        D[..., 2, 2] = self.E_f

        s_n1 = eps_n1[..., 1]

        tau_pi_trial = self.E_b * (s_n1 - xs_pi)
        X = self.gamma * alpha
        Z = self.K * z

        f_trial = np.fabs(tau_pi_trial - X) - self.tau_pi_bar - \
            Z + self.a * self.pressure / 3

        I = np.where(f_trial > 1e-6)

        sig = np.einsum('...st,...t->...s', D, eps_n1)
        sig[..., 1] = tau_pi_trial

        omega_I = omega[I]

        # Return mapping
        delta_lamda_I = f_trial[I] / \
            (self.E_b / (1 - omega_I) + self.gamma + self.K)

        # update all the state variables
        xs_pi[I] += (delta_lamda_I *
                     np.sign(tau_pi_trial[I] - X[I]) / (1 - omega_I))
        z[I] += delta_lamda_I
        alpha[I] += delta_lamda_I * np.sign(tau_pi_trial[I] - X[I])

        Y_I = 0.5 * self.E_b * (s_n1[I] - xs_pi[I]) ** 2
        omega[I] += (
            delta_lamda_I *
            (1 - omega_I) ** self.c *
            (Y_I / self.S) ** self.r
        )
        sig[..., 1] = (1 - omega) * self.E_b * (s_n1 - xs_pi)
        omega_I = omega[I]
        O = np.where(np.fabs(1. - omega_I) > 1e-5)
        IO = tuple([I[o][O] for o in range(len(I))])
        omega_IO = omega[IO]
        D_ed_IO = (
            self.E_b * (1 - omega_IO) - ((1 - omega_IO) * self.E_b **
                                         2) / (self.E_b + (self.gamma + self.K) * (1 - omega_IO))
            - ((1 - omega_IO) ** self.c * (self.E_b ** 2) * ((Y_I[O] / self.S) ** self.r)
               * np.sign(tau_pi_trial[I][O] - X[I][O]) * (s_n1[I][O] - xs_pi[I][O])) / ((self.E_b / (1 - omega_IO)) + self.gamma + self.K)
        )
        D[..., 1, 1] = (1 - omega) * self.E_b
        IO11 = IO + (1, 1)
        D[IO11] = D_ed_IO

        return sig, D

    ipw_view = bu.View(
        bu.Item('E_m'),
        bu.Item('E_f'),
        bu.Item('E_b'),
        bu.Item('tau_pi_bar'),
        bu.Item('gamma'),
        bu.Item('K'),
        bu.Item('S'),
        bu.Item('r'),
        bu.Item('c'),
        bu.Item('pressure'),
        bu.Item('a'),
    )



'''
Created on Dec 5, 2018

@author: rch
'''

import numpy as np
from ibvpy.view.plot2d import Viz2D


class Viz2DSigEps(Viz2D):

    def plot(self, ax, vot=1.0):
        prim = self.vis2d
        U = np.array(prim.U, np.float64)
        F = np.array(prim.F, np.float64)
        ax.plot(U[:, 0], F[:, 0])
        ax.plot(U[:, 1], F[:, 0])



'''
Created on Dec 17, 2018

@author: rch
'''

from traits.api import Interface, Type


class ITModel(Interface):
    r'''
    Interface of a time-stepping model capturing the state
    of a modeled object at the particular instance of time 
    supplying the predictor and corrector interface.

    There are three categories of parameters 

    * model parameters capturing the

      * geometry
      * boundary conditions
      * material characteristics
      * type of material behavior

    * control variable u that is associated with 
      state ``s`` 

    * for the given state response variables ``r(u,s)`` and eventually
      their derivatives 

    To support an iterative scheme with prediction based 
    on discrete the model can store several time steps  
    needed to make predictions for the next step.

    A model communicates with the time loop through their 
    common interface which depends on the particular type 
    of time-stepping scheme. In case of a time integration 
    scheme the control variable is represented by displacements,
    in case of an optimization problem, the function supplying 
    the values and gradients of response variables is required.
    In case of fragmentation problem, the model supplies 
    the array of load factors need to induce a crack in 
    a material point. 
    '''
    tloop_type = Type

    tstep_type = Type

    def get_corr_pred(self, U_k, t_n):
        '''Return the value and the derivative of a function
        '''

    def init_state(self):
        '''Initialize state.
        '''

    def get_state(self):
        '''Get the current control and primary variables.
        '''

    def update_state(self, U_k, t_n):
        '''Update the control, primary and state variables..
        '''

    def record_state(self):
        '''Provide the current state for history recording.
        '''

    def get_R(self, u, t):
        '''Get the value of the residuum.
        '''

    def get_dR(self, u, t):
        '''Get the gradient.
        '''





from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets2D.fets2D4q8u import FETS2D4Q8U
from ibvpy.api import\
     BCSlice, TStepper as TS, TLoop, TLine, RTDofGraph
from ibvpy.rtrace.rt_domain_list_field import RTraceDomainListField
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
from ibvpy.mesh.fe_domain import FEDomain

def combined_fe2D4q_with_fe2D4q8u():

    fets_eval_4u_conc = FETS2D4Q( mats_eval = MATS2DElastic( E = 28500, nu = 0.2 ) )
    fets_eval_4u_steel = FETS2D4Q( mats_eval = MATS2DElastic( E = 210000, nu = 0.25 ) )
    fets_eval_8u = FETS2D4Q8U( mats_eval = MATS2DElastic() )

    # Discretization
    fe_domain = FEDomain()

    fe_grid_level1 = FERefinementGrid( name = 'master grid',
                                       fets_eval = fets_eval_4u_conc,
                                       domain = fe_domain )

    fe_grid = FEGrid( level = fe_grid_level1,
                      coord_max = ( 2., 6., 0. ),
                      shape = ( 11, 30 ),
                      fets_eval = fets_eval_4u_conc )

    fe_grid_level2 = FERefinementGrid( name = 'refinement grid',
                                       parent = fe_grid_level1,
                                       fets_eval = fets_eval_4u_steel,
                                       fine_cell_shape = ( 1, 1 ) )

    # fe_grid_level1[ 5, :5 ].refine_using( fe_grid_level2 )
    # 1. first get the slice for the level - distinguish it from the slice at the subgrid
    #    this includes slicing in the subgrids. what if the subgrid does not exist?
    #    
    #    Each subgrid must hold its own slice within the level. The index operator fills
    #    the grid [...] instanciates the whole grid and returns the instance of 
    #    FEGridLevelSlice. The expanded subgrid contains its constructor slice.
    #
    # 2. If the slice is within an existing slice no change in the FESubgrid is required
    #    only the instance of the slice is returned. The FEGridLevelSlice goes always into 
    #    an expanded part of FEGrid.
    #
    # 3. If the slice does not fit into any existing slice - all domain with an intersection
    #    of the existing slice must be constructed as well. 
    #
    # 2. deactivate elements
    # 3.
    # BUT how to impose the boundary conditions on the particular refinement? The
    # slice has an attribute  

    fe_grid_level2.refine_elem( ( 5, 0 ) )
    fe_grid_level2.refine_elem( ( 5, 1 ) )
    fe_grid_level2.refine_elem( ( 5, 2 ) )
    fe_grid_level2.refine_elem( ( 5, 3 ) )
    fe_grid_level2.refine_elem( ( 5, 4 ) )
    fe_grid_level2.refine_elem( ( 5, 5 ) )

    # apply the boundary condition on a subgrid
    #
    print(fe_grid_level2.fe_subgrids)
    fe_first_grid = fe_grid_level2.fe_subgrids[0]

    ts = TS( dof_resultants = True,
             sdomain = fe_domain,
             bcond_list = [BCSlice( var = 'f', value = 1., dims = [0],
                                       slice = fe_grid[ :, -1, :, -1 ] ),
                           BCSlice( var = 'u', value = 0., dims = [0, 1],
                                       slice = fe_first_grid[ :, 0, :, 0 ] )
                                       ],
             rtrace_list = [ RTDofGraph( name = 'Fi,right over u_right (iteration)' ,
                                   var_y = 'F_int', idx_y = 0,
                                   var_x = 'U_k', idx_x = 1 ),
                        RTraceDomainListField( name = 'Stress',
                             var = 'sig_app', idx = 0, warp = True ),
#                             RTraceDomainField(name = 'Displacement' ,
#                                        var = 'u', idx = 0),
#                                 RTraceDomainField(name = 'N0' ,
#                                              var = 'N_mtx', idx = 0,
#                                              record_on = 'update')
                    ]
                )

    # Add the time-loop control
    tloop = TLoop( tstepper = ts,
                   tline = TLine( min = 0.0, step = 1, max = 1.0 ) )

    print(tloop.eval())
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = tloop )
    ibvpy_app.main()

#    print ts.F_int
#    print ts.rtrace_list[0].trace.ydata

if __name__ == '__main__':
    combined_fe2D4q_with_fe2D4q8u()




from traits.api import HasTraits, WeakRef
from ibvpy.view.ui.bmcs_tree_node import BMCSTreeNode
import bmcs_utils.api as bu
import numpy as np
import sympy as sp

class StrainNorm2D(BMCSTreeNode):

    mats = WeakRef

    def get_eps_eq(self, eps_Emef, kappa_Em):
        raise NotImplementedError

    def get_deps_eq(self, eps_Emef):
        raise NotImplementedError

    def update_plot(self, ax):
        n_i = 100
        eps_i = np.linspace(-3, 3, n_i)
        eps_11, eps_22 = np.meshgrid(eps_i, eps_i)
        eps_11_ = eps_11.flatten()
        eps_22_ = eps_22.flatten()
        eps_ief = np.zeros((len(eps_11_), 2, 2))
        eps_ief[:, 0, 0] = eps_11_
        eps_ief[:, 1, 1] = eps_22_
        kappa_i = np.zeros_like(eps_11_)
        eps_eq_i = self.get_eps_eq(eps_ief, kappa_i)
        eps_eq_ij = eps_eq_i.reshape(n_i,n_i)
        ax.contour(eps_11, eps_22, eps_eq_ij) # , [1]) # , label=r'$\kappa = 1$')
        ax.plot([0,0],[-3,3],color='black', lw=0.3)
        ax.plot([-3,3],[0,0],color='black', lw=0.3)
        ax.set_xlabel(r'$\varepsilon_{1}$')
        ax.set_ylabel(r'$\varepsilon_{2}$')
        ax.axis('equal')

class SN2DRankine(StrainNorm2D):
    '''
    Computes principal strains and makes a norm of their positive part
    '''
    name = 'Rankine strain norm'

    def get_eps_eq(self, eps_Emef, kappa_Em):

        eps_11 = eps_Emef[..., 0, 0]
        eps_22 = eps_Emef[..., 1, 1]
        eps_12 = eps_Emef[..., 0, 1]
        eps_eq_Em = (
            0.5 * (eps_11 + eps_22) +
            np.sqrt(((eps_11 - eps_22) / 2.0)**2.0 + eps_12**2.0)
        )
        e_Em = np.concatenate(
            (eps_eq_Em[..., None], kappa_Em[..., None]), axis=-1
        )
        eps_eq = np.max(e_Em, axis=-1)
        return eps_eq

    def get_deps_eq(self, eps_Emef):
        eps11 = eps_Emef[..., 0, 0]
        eps22 = eps_Emef[..., 1, 1]
        eps12 = eps_Emef[..., 0, 1]
        eps_11_22 = eps11 - eps22

        denom = 2. * np.sqrt(eps_11_22 * eps_11_22 + 4.0 * eps12 * eps12)
        factor = np.zeros_like(denom)
        nz_idx = np.where(denom != 0.0)
        factor[nz_idx] = 1. / denom[nz_idx]
        df_trial1 = factor * np.array([[eps11 - eps22, 4.0 * eps12],
                                       [4.0 * eps12, eps22 - eps11]])
        return (np.einsum('ab...->...ab', df_trial1) +
                0.5 * np.identity(2)[None, :, :])

class SN2MasarsExpr(bu.SymbExpr):

    eps_11, eps_12, eps_22 = sp.symbols(
        r'\varepsilon_{11}, \varepsilon_{22}, \varepsilon_{12}')
    eps_tns = sp.Matrix([[eps_11, eps_12],[eps_12, eps_22]])
    eps_1, eps_2 = eps_tns.eigenvals()
    def pos(x):
        return sp.Piecewise((0, x<0),
                            (x, True))
    eps_1_pos, eps_2_pos = pos(eps_1), pos(eps_2)
    eps_eq = sp.simplify(sp.sqrt( eps_1_pos**2 + eps_2_pos**2))

    d_eps_eq_11 = eps_eq.diff(eps_11)
    d_eps_eq_12 = eps_eq.diff(eps_12)
    d_eps_eq_22 = eps_eq.diff(eps_22)

    symb_model_params = []

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('eps_eq', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_11', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_12', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_22', ('eps_11', 'eps_22', 'eps_12')),
    ]

class SN2DMasars(StrainNorm2D,bu.InjectSymbExpr):
    '''
    Computes principal strains and makes a norm of their positive part
    '''
    symb_class = SN2MasarsExpr
    name = 'Masars strain norm'

    def get_eps_eq(self, eps_Emef, kappa_Em):
        eps_11 = eps_Emef[..., 0, 0]
        eps_22 = eps_Emef[..., 1, 1]
        eps_12 = eps_Emef[..., 0, 1]
        eps_eq_Em = self.symb.get_eps_eq(eps_11, eps_22, eps_12)
        e_Em = np.concatenate(
            (eps_eq_Em[..., None], kappa_Em[..., None]), axis=-1
        )
        eps_eq = np.max(e_Em, axis=-1)
        return eps_eq

    def get_deps_eq(self, eps_Emef):
        eps_11 = eps_Emef[..., 0, 0]
        eps_22 = eps_Emef[..., 1, 1]
        eps_12 = eps_Emef[..., 0, 1]
        d_eps_eq_11 = self.symb.get_d_eps_eq_11(eps_11, eps_22, eps_12)
        d_eps_eq_22 = self.symb.get_d_eps_eq_22(eps_11, eps_22, eps_12)
        d_eps_eq_12 = self.symb.get_d_eps_eq_12(eps_11, eps_22, eps_12)
        d_eps_eq_efEm = np.array([[d_eps_eq_11, d_eps_eq_12],
                                  [d_eps_eq_12, d_eps_eq_22]])
        return np.einsum('ef...->...ef', d_eps_eq_efEm)

class SN2EnergyExpr(bu.SymbExpr):

    eps_11, eps_12, eps_22 = sp.symbols(
        r'\varepsilon_{11}, \varepsilon_{22}, \varepsilon_{12}')
    eps_tns = sp.Matrix([[eps_11, eps_12],[eps_12, eps_22]])
    eps_1, eps_2 = eps_tns.eigenvals()
    def pos(x):
        return sp.Piecewise((0, x<0),
                            (x, True))
    eps_1_pos, eps_2_pos = pos(eps_1), pos(eps_2)
    eps_eq = sp.simplify(sp.sqrt( eps_1_pos**2 + eps_2_pos**2))

    d_eps_eq_11 = eps_eq.diff(eps_11)
    d_eps_eq_12 = eps_eq.diff(eps_12)
    d_eps_eq_22 = eps_eq.diff(eps_22)

    symb_model_params = []

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('eps_eq', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_11', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_12', ('eps_11', 'eps_22', 'eps_12')),
        ('d_eps_eq_22', ('eps_11', 'eps_22', 'eps_12')),
    ]

class SN2DEnergy(StrainNorm2D,bu.InjectSymbExpr):
    '''
    Computes the energy norm

    \kappa = \sqrt{ \varepsilon_{ab} D_{abcd} \varepsilon_{cd}}

    the derivative ...
    \dfrac{\partial \kappa}{\partial \varepsilon_{cd}}
    =
    \dfrac{1}{\kappa} D_{abcd} \varepsilon_{cd}
    '''
    symb_class = SN2EnergyExpr
    name = 'Energy norm'

    def get_eps_eq(self, eps_Emef, kappa_Em):
        D_abcd = self.mats.D_abcd
        eps_eq_Em = np.sqrt( np.einsum('...ab,abcd,...cd->...', eps_Emef, D_abcd, eps_Emef))
        e_Em = np.concatenate(
            (eps_eq_Em[..., None], kappa_Em[..., None]), axis=-1
        )
        eps_eq = np.max(e_Em, axis=-1)
        return eps_eq

    def get_deps_eq(self, eps_Emef):
        D_abcd = self.mats.D_abcd
        eps_eq_Em = np.sqrt( np.einsum('...ab,abcd,...cd->...', eps_Emef, D_abcd, eps_Emef))
        deps_eq = np.einsum('...,...abcd,...cd->...ab', 1 / eps_eq_Em, D_abcd, eps_Emef)
        return deps_eq



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 7, 2009 by: rchx
'''
A trait type with shadow attribute that allows database referencing -
database keys are allowed as values, the shadow attribute is then the
appropriate database member.
'''

from traits.api import \
    TraitType, HasStrictTraits, TraitError, \
    Property, Button

class KeyRef(TraitType):

    is_mapped = True

    def __init__(self, *args, **metadata):
        if len(args) == 1:
            self._default_key = args[0]
        else:
            self._default_key = None
        db = metadata.get('db', None)
        if db:
            self.map = db
        else:
            raise ValueError('db not defined')

        super(KeyRef, self).__init__(**metadata)

    def validate(self, obj, name, key):
        ''' Set the trait value '''

        self.keys_name = name + '_keys'
        keys_prop = Property(fget=lambda: list(self.map.keys()))
        obj.add_trait(self.keys_name, keys_prop)

        if key in list(self.map.keys()):
            prechange_mapped_val = getattr(obj, name + '_')
            try:
                prechange_mapped_val.del_link(obj)
                '''Management of backward links - erase the link to an object that has
                stopped using me.
                '''
            except AttributeError:
                print(('Warning: Unable to remove reference to object from '
                'mapped value of type %s (Backwards link management not implemented'
                ' in mapped type), trait name %s') % (prechange_mapped_val.__class__, name))
            postchange_mapped_val = self.mapped_value(key)
            try:
                postchange_mapped_val.add_link(obj)
                '''Management of backward links - add the link to an object that has
                started using me.
                '''
            except AttributeError:
                print(('Warning: Unable to pass reference to object to mapped '
                'value of type %s (Backwards link management not implemented '
                'in mapped type)') % prechange_mapped_val.__class__)

            self.post_setattr(obj, name, key)
            ''' When the new value is not different from the old,
            post_setattr does not get called, which can cause issues
            '''
            return key
        else:
            self.error(object, name, key)

    def get_default_value(self):
        '''Take the default value'''

        keys = list(self.map.keys())
        if self._default_key == None:
            if len(keys) > 1:
                return (0, list(self.map.keys())[0])
            else:
                return (0, None)
                raise TraitError('invalid key, no entries in database extension %s' % \
                    (self.map.klass,))
        if self._default_key in keys:
            return (0, self._default_key)

        raise TraitError('assigned default value must be one of %s but a value of %s was received' % \
            (list(self.map.keys()), self._default_key))

    def mapped_value (self, key):
        try:
            return self.map[ key ]
        except:
            return None

    def post_setattr (self, object, name, value):
        val = self.mapped_value(value)
        try:
            setattr(object, name + '_', val)
        except:
            # We don't need a fancy error message, because this exception
            # should always be caught by a TraitCompound handler:
            raise TraitError('Unmappable')

    def info (self):
        keys = [ repr(x) for x in list(self.map.keys()) ]
        keys.sort()
        return ' or '.join(keys)

    def get_editor (self, trait=None):
        return self.create_editor()

    def create_editor(self):
        pass




'''
Created on Feb 11, 2018

@author: rch
'''

import os

from mayavi import mlab
from mayavi.filters.api import ExtractTensorComponents
from mayavi.modules.api import Surface
from mayavi.sources.vtk_xml_file_reader import VTKXMLFileReader
from tvtk.api import write_data

import numpy as np
import traits.api as tr

from .viz3d_field import Vis3DField, Viz3DField


class Vis3DTensorField(Vis3DField):

    var = tr.Str('eps_ab')

    def update(self):
        #ts = self.sim.tstep
        ts = self.tstep
        U = ts.U_k
        t_n1 = ts.t_n1
        U_vector_fields = []
        var_Encd_tensor_fields = []
        for domain in ts.fe_domain:
            xdomain = domain.xmodel
            var_function = domain.tmodel.var_dict.get(self.var, None)
            if var_function == None or xdomain.hidden:
                continue
            fets = xdomain.fets
            DELTA_x_ab = fets.vtk_expand_operator
            DELTA_f_ab = xdomain.vtk_expand_operator
            U_Eia = U[xdomain.o_Eia]
            _, _, n_a = U_Eia.shape
            U_vector_fields.append(np.einsum(
                'Ia,ab->Ib', U_Eia.reshape(-1, n_a), DELTA_x_ab
            ))
            eps_Enab = xdomain.map_U_to_field(U)
            # copy the state variable as the operator
            var_Enab = var_function(eps_Enab, t_n1, **domain.state_k)
            var_Encd = np.einsum(
                '...ab,ac,bd->...cd',
                var_Enab, DELTA_f_ab, DELTA_f_ab
            )
            var_Encd_tensor_fields.append(var_Encd.reshape(-1, 9))
        self.ug.point_data.vectors = np.vstack(U_vector_fields)
        self.ug.point_data.vectors.name = 'displacement'
        self.ug.point_data.tensors = np.vstack(var_Encd_tensor_fields)
        self.ug.point_data.tensors.name = self.var
        fname = '%s_step_%08.4f' % (self.var, t_n1)
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.' + self.extension
        write_data(self.ug, target_file)
        self.add_file(target_file)


class Viz3DTensorField(Viz3DField):

    def setup(self):
        m = mlab
        fname = self.vis3d.file_list[0]
        var = self.vis3d.var
        self.d = VTKXMLFileReader()
        self.d.initialize(fname)
        self.src = m.pipeline.add_dataset(self.d)
        self.warp_vector = m.pipeline.warp_vector(self.src)
        #self.surf = m.pipeline.surface(self.warp_vector)
        engine = m.get_engine()
        self.etc = ExtractTensorComponents()
        engine.add_filter(self.etc, self.warp_vector)
        surface2 = Surface()
        self.surface = surface2
        engine.add_filter(surface2, self.etc)
        lut = self.etc.children[0]
        lut.scalar_lut_manager.set(
            show_scalar_bar=True,
            show_legend=True,
            data_name=var
        )
        lut.scalar_lut_manager.scalar_bar.orientation = 'horizontal'
        lut.scalar_lut_manager.scalar_bar_representation.trait_set(
            maximum_size=np.array([100000, 100000]),
            minimum_size=np.array([1, 1]),
            position=np.array([0.3, 0.05]),
            position2=np.array([0.65, 0.1]),
        )

        lut.scalar_lut_manager.label_text_property.trait_set(
            font_family='times',
            italic=False,
            bold=False
        )
        lut.scalar_lut_manager.title_text_property.trait_set(
            font_family='times',
            italic=False,
            bold=False
        )
        self.etc.filter.scalar_mode = 'component'





from traits.api import \
    Interface, Tuple


class IXModel(Interface):
    '''Mappings between spatial domain representation 
    and between the linear algebr objects. 
    '''
    U_var_shape = Tuple

    state_var_shape = Tuple

    def map_U_to_field(self, eps_eng): pass

    def map_field_to_F(self, eps_tns): pass

    def map_field_to_K(self, tns4): pass



'''
Created on Mar 16, 2017

@author: rch
'''

import numpy as np
from .viz2d import Viz2D


class Viz2DTimeFunction(Viz2D):

    def plot(self, ax, vot):
        x, y = self.vis2d.get_viz2d_data()
        ax.plot(x, y)
        ymin, ymax = np.min(y), np.max(y)
        ax.plot([vot, vot], [ymin, ymax])



'''
Created on Feb 14, 2019

@author: rch
'''

from ibvpy.tmodel import MATSEval
from ibvpy.sim.i_tmodel import ITModel
from traits.api import on_trait_change

from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, LiDamageFn, ExpSlopeDamageFn, AbaqusDamageFn, \
    MultilinearDamageFn, \
    FRPDamageFn
import numpy as np
import traits.api as tr


@tr.provides(ITModel)
class MATS1D5DP(MATSEval):

    node_name = 'Pressure Sensitive bond model'

    E_N = tr.Float(100, label='E_N',
                   desc='Normal stiffness of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    E_T = tr.Float(100, label='E_T',
                   desc='Shear modulus of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    gamma = tr.Float(40.0, label='gamma',
                     desc='Kinematic Hardening Modulus',
                     MAT=True,
                     enter_set=True, auto_set=False)

    K = tr.Float(1, label='K',
                 desc='Isotropic hardening modulus',
                 MAT=True,
                 enter_set=True, auto_set=False)

    _c = tr.Float(1, Label='c',
                  desc='Damage accumulation parameter',
                  MAT=True,
                  enter_set=True, auto_set=False)

    tau_bar = tr.Float(1, label='tau_bar',
                       desc='Reversibility limit',
                       MAT=True,
                       enter_set=True, auto_set=False)

    state_var_shapes = dict(s_p=(),
                            alpha=(),
                            z=(),
                            omega=(),
                            kappa=())

    uncoupled_dp = tr.Bool(False,
                           MAT=True,
                           label='Uncoupled d-p'
                           )

    s_0 = tr.Float(0.01, MAT=True,
                   desc='Elastic strain/displacement limit')

    def __init__(self, *args, **kw):
        super(MATS1D5DP, self).__init__(*args, **kw)
        self._update_s0()

    @on_trait_change('tau_bar,E_T')
    def _update_s0(self):
        if not self.uncoupled_dp:
            if self.E_T == 0:
                self.s_0 = 0
            else:
                self.s_0 = self.tau_bar / self.E_T
            self.omega_fn.s_0 = self.s_0

    omega_fn_type = tr.Trait('FRP',
                             dict(li=LiDamageFn,
                                  jirasek=ExpSlopeDamageFn,
                                  abaqus=AbaqusDamageFn,
                                  FRP=FRPDamageFn,
                                  multilinear=MultilinearDamageFn
                                  ),
                             MAT=True,
                             )

    @on_trait_change('omega_fn_type')
    def _reset_omega_fn(self):
        print('resetting')
        self.omega_fn = self.omega_fn_type_(s_0=self.s_0)

    omega_fn = tr.Instance(IDamageFn, report=True)

    def _omega_fn_default(self):
        return MultilinearDamageFn()

    def omega(self, k):
        return self.omega_fn(k)

    def omega_derivative(self, k):
        return self.omega_fn.diff(k)

    def init(self, s_pi, alpha, z, omega, kappa):
        r'''
        Initialize the state variables.
        '''
        s_pi[...] = 0
        alpha[...] = 0
        z[...] = 0
        omega[...] = 0
        kappa[...] = 0

    algorithmic = tr.Bool(True)

    def get_corr_pred(self, u_r, t_n, s_p, alpha, z, omega, kappa):

        s = u_r[..., 0]
        w = u_r[..., 1]
        # For normal
        H_w_N = np.array(w <= 0.0, dtype=np.float64)
        E_alg_N = H_w_N * self.E_N
        sig_N = E_alg_N * w
        sig_pi_trial = self.E_T * (s - s_p)
        Z = self.K * z
        # for handling the negative values of isotropic hardening
        h_1 = self.tau_bar + Z
        pos_iso = h_1 > 1e-6

        X = self.gamma * alpha

        # for handling the negative values of kinematic hardening (not yet)
        # h_2 = h * np.sign(sig_pi_trial - X) * \
        #    np.sign(sig_pi_trial) + X * np.sign(sig_pi_trial)
        #pos_kin = h_2 > 1e-6

        f = np.fabs(sig_pi_trial - X) - h_1 * pos_iso

        I = f > 1e-6
        # Return mapping
        delta_lamda_I = f[I] / (self.E_T + self.gamma + np.fabs(self.K))
        # update all the state variables
        s_p[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])
        z[I] += delta_lamda_I
        alpha[I] += delta_lamda_I * np.sign(sig_pi_trial[I] - X[I])

        kappa[...] = np.max(np.array([kappa, np.fabs(s)]), axis=0)
        omega[...] = self.omega(kappa)
        tau = (1 - omega) * self.E_T * (s - s_p)

        E_alg_T = (1 - omega) * self.E_T

        domega_ds = self.omega_derivative(kappa)
        # Consistent tangent operator
        E_alg_T[I] = -self.E_T / (self.E_T + self.K + self.gamma) \
            * domega_ds[I] * self.E_T * (s[I] - s_p[I]) \
            + (1 - omega[I]) * self.E_T * (self.K + self.gamma) / \
            (self.E_T + self.K + self.gamma)

        sig = np.zeros_like(u_r)
        sig[..., 0] = tau
        sig[..., 1] = sig_N
        E_TN = np.einsum('abEm->Emab',
                         np.array(
                             [
                                 [E_alg_T, np.zeros_like(E_alg_T)],
                                 [np.zeros_like(E_alg_N), E_alg_N]
                             ])
                         )
        return sig, E_TN
